# Copyright 2019 Okera Inc. All Rights Reserved.
#
# Some integration tests for managing attributes
#
# pylint: disable=bad-continuation
# pylint: disable=line-too-long
# pylint: disable=too-many-locals
# pylint: disable=too-many-lines

import unittest

from okera.tests import pycerebro_test_common as common

class AttributesTest(unittest.TestCase):

    # skip_cascade is a flag to skip tests for known bugs, remove when fixed
    def _validate(self, conn, db, attr, tags, sql, expected, skip_cascade=False):
        # Create the view and then assign the tag, validate tag INHERITANCE
        conn.execute_ddl("DROP VIEW IF EXISTS %s.v" % db)
        conn.execute_ddl("CREATE VIEW %s.v AS %s" % (db, sql))

        conn.execute_ddl("drop attribute if exists %s.a1" % db)
        conn.execute_ddl("create attribute %s.a1" % db)
        for tag in tags:
            conn.assign_attribute(db, attr, tag[0], tag[1], tag[2], cascade=True)

        print(conn.execute_ddl_table_output('DESCRIBE %s.v' % db))
        self.assertEqual(
            expected,
            str(conn.execute_ddl_table_output('DESCRIBE %s.v' % db)))

        # Recreate the view (tagged already assigned), validate tag CASCADE
        if not skip_cascade:
            conn.execute_ddl("DROP VIEW IF EXISTS %s.v" % db)
            conn.execute_ddl("CREATE VIEW %s.v AS %s" % (db, sql))
            print(conn.execute_ddl_table_output('DESCRIBE %s.v' % db))
            self.assertEqual(
                expected,
                str(conn.execute_ddl_table_output('DESCRIBE %s.v' % db)))

    def test_specific(self):
        # Not a test, just skeleton to test a specific scenario
        db = 'attr_test_db'
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            conn.execute_ddl("drop database if exists %s cascade" % db)
            conn.execute_ddl("create database if not exists %s" % db)

            self._validate(conn, db, 'a1',
                [['chase', 'subscription_currency', 'currency.country']],
                'select ** from chase.subscription_currency',
                """
+------------------------+--------+---------+-----------------+
|          name          |  type  | comment |    attributes   |
+------------------------+--------+---------+-----------------+
|   currency__country    | string |         | attr_test_db.a1 |
| currency__currencycode | string |         |                 |
+------------------------+--------+---------+-----------------+""".strip(), skip_cascade=True)

    def test_unnest_attributes(self):
        db = 'attr_test_db'
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            conn.execute_ddl("drop database if exists %s cascade" % db)
            conn.execute_ddl("create database if not exists %s" % db)

            # No tags
            self._validate(conn, db, 'a1', [],
                """select productkey, partyroles.item.partykey as partykey
                   from chase.zd1238_4, chase.zd1238_4.partyroles""",
                """
+------------+--------+---------+------------+
|    name    |  type  | comment | attributes |
+------------+--------+---------+------------+
| productkey | string |         |            |
|  partykey  | string |         |            |
+------------+--------+---------+------------+""".strip())

            self._validate(conn, db, 'a1',
                [['chase', 'zd1238_4', 'productkey'],
                 ['chase', 'zd1238_4', 'partyroles.partykey']],
                """select productkey, partyroles.item.partykey as partykey
                   from chase.zd1238_4, chase.zd1238_4.partyroles""",
                """
+------------+--------+---------+-----------------+
|    name    |  type  | comment |    attributes   |
+------------+--------+---------+-----------------+
| productkey | string |         | attr_test_db.a1 |
|  partykey  | string |         | attr_test_db.a1 |
+------------+--------+---------+-----------------+""".strip())

            self._validate(conn, db, 'a1',
                [['chase', 'zd1238_4', 'productkey']],
                """select productkey, partyroles.item.partykey as partykey
                   from chase.zd1238_4, chase.zd1238_4.partyroles""",
                """
+------------+--------+---------+-----------------+
|    name    |  type  | comment |    attributes   |
+------------+--------+---------+-----------------+
| productkey | string |         | attr_test_db.a1 |
|  partykey  | string |         |                 |
+------------+--------+---------+-----------------+""".strip())

            self._validate(conn, db, 'a1',
                [['chase', 'zd1238_4', 'partyroles.partykey']],
                """select productkey, partyroles.item.partykey as partykey
                   from chase.zd1238_4, chase.zd1238_4.partyroles""",
                """
+------------+--------+---------+-----------------+
|    name    |  type  | comment |    attributes   |
+------------+--------+---------+-----------------+
| productkey | string |         |                 |
|  partykey  | string |         | attr_test_db.a1 |
+------------+--------+---------+-----------------+""".strip())

            self._validate(conn, db, 'a1',
                [['rs_complex', 'strarray_t', 'str_arr']],
                """select id, str_arr.item
                   from rs_complex.strarray_t, rs_complex.strarray_t.str_arr""",
                """
+------+--------+---------+-----------------+
| name |  type  | comment |    attributes   |
+------+--------+---------+-----------------+
|  id  | bigint |         |                 |
| item | string |         | attr_test_db.a1 |
+------+--------+---------+-----------------+""".strip())

            self._validate(conn, db, 'a1',
                [['rs_complex', 'array_struct_t', 'a1']],
                """select unnest_alias.item, unnest_alias.item.f1 as f1
                   from rs_complex.array_struct_t,
                   rs_complex.array_struct_t.a1 unnest_alias""",
                """
+------+--------------+---------+-----------------+
| name |     type     | comment |    attributes   |
+------+--------------+---------+-----------------+
| item |   struct<    |         | attr_test_db.a1 |
|      |   f1:string, |         |                 |
|      |   f2:string  |         |                 |
|      |      >       |         |                 |
|  f1  |    string    |         |                 |
+------+--------------+---------+-----------------+""".strip())

            self._validate(conn, db, 'a1',
                [['chase', 'subscription_currency', 'currency.country']],
                'select ** from chase.subscription_currency',
                """
+------------------------+--------+---------+-----------------+
|          name          |  type  | comment |    attributes   |
+------------------------+--------+---------+-----------------+
|   currency__country    | string |         | attr_test_db.a1 |
| currency__currencycode | string |         |                 |
+------------------------+--------+---------+-----------------+""".strip())

            # FIXME: this is not cascading correctly
            self._validate(conn, db, 'a1',
                [['rs_complex', 'array_struct_t', 'a1.f1']],
                """select unnest_alias.item, unnest_alias.item.f1 as f1
                   from rs_complex.array_struct_t,
                   rs_complex.array_struct_t.a1 unnest_alias""",
                """
+------+--------------+---------+-----------------+
| name |     type     | comment |    attributes   |
+------+--------------+---------+-----------------+
| item |   struct<    |         |                 |
|      |   f1:string, |         | attr_test_db.a1 |
|      |   f2:string  |         |                 |
|      |      >       |         |                 |
|  f1  |    string    |         | attr_test_db.a1 |
+------+--------------+---------+-----------------+""".strip(), skip_cascade=True)

            self._validate(conn, db, 'a1',
                [['rs_complex', 'array_struct_t', 'a1.f2']],
                """select a1.item.f2 as f2, a1.item.f1 as f1
                   from rs_complex.array_struct_t, rs_complex.array_struct_t.a1""",
                """
+------+--------+---------+-----------------+
| name |  type  | comment |    attributes   |
+------+--------+---------+-----------------+
|  f2  | string |         | attr_test_db.a1 |
|  f1  | string |         |                 |
+------+--------+---------+-----------------+""".strip())

            self._validate(conn, db, 'a1',
                [['rs_complex', 'array_struct_t', 'a1.f2'],
                 ['rs_complex', 'array_struct_t', 'a1.f1']],
                """select a1.item.f2 as f2, a1.item.f1 as f1
                   from rs_complex.array_struct_t, rs_complex.array_struct_t.a1""",
                """
+------+--------+---------+-----------------+
| name |  type  | comment |    attributes   |
+------+--------+---------+-----------------+
|  f2  | string |         | attr_test_db.a1 |
|  f1  | string |         | attr_test_db.a1 |
+------+--------+---------+-----------------+""".strip())

            # FIXME: this is not cascading correctly
            self._validate(conn, db, 'a1',
               [['rs_complex', 'array_struct_t', 'a1.f1'],
                 ['rs_complex', 'array_struct_t', 'a1.f2']],
                """select a1.item as item, a1.item.f1 as f1
                   from rs_complex.array_struct_t, rs_complex.array_struct_t.a1""",
                """
+------+--------------+---------+-----------------+
| name |     type     | comment |    attributes   |
+------+--------------+---------+-----------------+
| item |   struct<    |         |                 |
|      |   f1:string, |         | attr_test_db.a1 |
|      |   f2:string  |         | attr_test_db.a1 |
|      |      >       |         |                 |
|  f1  |    string    |         | attr_test_db.a1 |
+------+--------------+---------+-----------------+""".strip(), skip_cascade=True)

            self._validate(conn, db, 'a1',
                [['rs_complex', 'array_struct_array', 'a1.a2'],
                 ['rs_complex', 'array_struct_array', 'a1.f1']],
                """select a1.item.a2 as a2, a1.item.f1 as f1
                   from rs_complex.array_struct_array,
                   rs_complex.array_struct_array.a1""",
                """
+------+---------------+---------+-----------------+
| name |      type     | comment |    attributes   |
+------+---------------+---------+-----------------+
|  a2  | array<string> |         | attr_test_db.a1 |
|  f1  |     string    |         | attr_test_db.a1 |
+------+---------------+---------+-----------------+""".strip())

            self._validate(conn, db, 'a1',
                [['chase', 'zd1238_4', 'partyroles.partykey']],
                """select partyroles.item.partykey as partykey,
                          tokenize(partyroles.item.partykey) as tokenized_key
                   from chase.zd1238_4 c, c.partyroles
                   where partyroles.item.partykey =
                      '1a43fd68-31d0-46a4-b3f0-bc42730ed5f7'""",
                """
+---------------+--------+---------+-----------------+
|      name     |  type  | comment |    attributes   |
+---------------+--------+---------+-----------------+
|    partykey   | string |         | attr_test_db.a1 |
| tokenized_key | string |         |                 |
+---------------+--------+---------+-----------------+""".strip())

            self._validate(conn, db, 'a1',
                [['rs_complex_parquet', 'spark_gzip', 'int32'],
                 ['rs_complex_parquet', 'spark_gzip', 'str_arr'],
                 ['rs_complex_parquet', 'spark_gzip', 'int_arr']],
                """select int32, str_arr.item as str_arr, int_arr.item as int_arr
                   from rs_complex_parquet.spark_gzip,
                        rs_complex_parquet.spark_gzip.str_arr,
                        rs_complex_parquet.spark_gzip.int_arr""",
                """
+---------+--------+---------+-----------------+
|   name  |  type  | comment |    attributes   |
+---------+--------+---------+-----------------+
|  int32  |  int   |         | attr_test_db.a1 |
| str_arr | string |         | attr_test_db.a1 |
| int_arr |  int   |         | attr_test_db.a1 |
+---------+--------+---------+-----------------+""".strip())

            self._validate(conn, db, 'a1',
                [['chase', 'zd1211', 'accountnumber'],
                 ['chase', 'zd1211', 'productdetails.subproducts']],
                """select accountnumber, subproducts.item
                   from chase.zd1211, chase.zd1211.productdetails.subproducts
                   where accountnumber in ('65360119', '26509759')""",
"""
+---------------+--------+---------+-----------------+
|      name     |  type  | comment |    attributes   |
+---------------+--------+---------+-----------------+
| accountnumber | string |         | attr_test_db.a1 |
|      item     | string |         | attr_test_db.a1 |
+---------------+--------+---------+-----------------+""".strip())

            self._validate(conn, db, 'a1',
                [['rs_complex', 'array_struct_array', 'a1.f1']],
                """select b.* from rs_complex.array_struct_array a, a.a1 b
                   where b.item.f1 = 'ab'""",
"""
+------+---------------+---------+-----------------+
| name |      type     | comment |    attributes   |
+------+---------------+---------+-----------------+
|  f1  |     string    |         | attr_test_db.a1 |
|  f2  |     string    |         |                 |
|  a2  | array<string> |         |                 |
+------+---------------+---------+-----------------+""".strip())

            self._validate(conn, db, 'a1',
                [['chase', 'product', 'productkey'],
                 ['chase', 'product', 'feescharges.feeamount'],
                 ['chase', 'product', 'feescharges.feename']],
                """select p.productkey, fc.item.feename as feename,
                   fc.item.feeamount as feeamount from chase.product p, p.feescharges fc
                   where fc.item.feeamount = '10.1' and fc.item.feename = 'STOPCHEQUE'""",
"""
+------------+--------+---------+-----------------+
|    name    |  type  | comment |    attributes   |
+------------+--------+---------+-----------------+
| productkey | string |         | attr_test_db.a1 |
|  feename   | string |         | attr_test_db.a1 |
| feeamount  | string |         | attr_test_db.a1 |
+------------+--------+---------+-----------------+""".strip())

            self._validate(conn, db, 'a1',
                [['chase', 'zd1211', 'accountnumber'],
                ['chase', 'zd1211', 'productdetails.subproducts']],
                """select accountnumber, subproducts.item from chase.zd1211,
                   chase.zd1211.productdetails.subproducts
                   where accountnumber in ('65360119', '26509759')""",
"""
+---------------+--------+---------+-----------------+
|      name     |  type  | comment |    attributes   |
+---------------+--------+---------+-----------------+
| accountnumber | string |         | attr_test_db.a1 |
|      item     | string |         | attr_test_db.a1 |
+---------------+--------+---------+-----------------+""".strip())

            self._validate(conn, db, 'a1',
                [['rs_complex_parquet', 'array_struct_t', 'a1'],
                 ['rs_complex_parquet', 'array_struct_t', 'a1.f1']],
                """select t1.item, t1.item.f1 as f1
                   from rs_complex_parquet.array_struct_t, rs_complex.array_struct_t,
                        rs_complex_parquet.array_struct_t.a1 t1
                   join rs_complex.array_struct_t.a1 t2 ON (t1.item.f1 = t2.item.f1)""",
"""
+------+--------------+---------+-----------------+
| name |     type     | comment |    attributes   |
+------+--------------+---------+-----------------+
| item |   struct<    |         | attr_test_db.a1 |
|      |   f1:string, |         | attr_test_db.a1 |
|      |   f2:string  |         |                 |
|      |      >       |         |                 |
|  f1  |    string    |         | attr_test_db.a1 |
+------+--------------+---------+-----------------+""".strip())

            self._validate(conn, db, 'a1',
                [['rs_complex_parquet', 'array_struct_t', 'a1.f1']],
                """select a1.* from rs_complex_parquet.array_struct_t,
                   rs_complex_parquet.array_struct_t.a1""",
"""
+------+--------+---------+-----------------+
| name |  type  | comment |    attributes   |
+------+--------+---------+-----------------+
|  f1  | string |         | attr_test_db.a1 |
|  f2  | string |         |                 |
+------+--------+---------+-----------------+""".strip())

            self._validate(conn, db, 'a1',
                [['rs_complex_parquet', 'array_struct_t', 'a1.f1'],
                 ['rs_complex_parquet', 'array_struct_t', 'a1.f2']],
                """select a1.* from rs_complex_parquet.array_struct_t,
                   rs_complex_parquet.array_struct_t.a1""",
"""
+------+--------+---------+-----------------+
| name |  type  | comment |    attributes   |
+------+--------+---------+-----------------+
|  f1  | string |         | attr_test_db.a1 |
|  f2  | string |         | attr_test_db.a1 |
+------+--------+---------+-----------------+""".strip())

            self._validate(conn, db, 'a1',
                [['chase', 'product', 'limits.transactionlimits.description'],
                 ['chase', 'product', 'limits.transactionlimits.transactionname']],
                """select tlimits.* from chase.product p,
                   p.limits.transactionlimits tlimits""",
"""
+-----------------+--------+---------+-----------------+
|       name      |  type  | comment |    attributes   |
+-----------------+--------+---------+-----------------+
| transactionname | string |         | attr_test_db.a1 |
|   description   | string |         | attr_test_db.a1 |
|  minimumamount  | string |         |                 |
|  maximumamount  | string |         |                 |
|   resetperiod   | string |         |                 |
+-----------------+--------+---------+-----------------+""".strip())

            #
            # FIXME: this is not right. Not inheriting inside f1. The cascade value
            # is actually correct.
            #
            self._validate(conn, db, 'a1',
                [['rs_complex', 'array_struct_array', 'a1'],
                 ['rs_complex', 'array_struct_array', 'a1.f1'],
                 ['rs_complex', 'array_struct_array', 'a1.a2']],
                """select a.*, b.* from rs_complex.array_struct_array a, a.a1 b""",
"""
+------+--------------------+---------+-----------------+
| name |        type        | comment |    attributes   |
+------+--------------------+---------+-----------------+
|  a1  |   array<struct<    |         | attr_test_db.a1 |
|      |      f1:string,    |         |                 |
|      |      f2:string,    |         |                 |
|      |   a2:array<string> |         |                 |
|      |         >>         |         |                 |
|  f1  |       string       |         | attr_test_db.a1 |
|  f2  |       string       |         |                 |
|  a2  |   array<string>    |         | attr_test_db.a1 |
+------+--------------------+---------+-----------------+""".strip(), skip_cascade=True)

            self._validate(conn, db, 'a1',
                [['functional', 'allcomplextypes', 'id'],
                 ['functional', 'allcomplextypes', 'nested_struct_col.f1']],
                """select * from functional.allcomplextypes,
                   functional.allcomplextypes.int_array_col""",
"""
+---------------------------+----------------------------+---------+-----------------+
|            name           |            type            | comment |    attributes   |
+---------------------------+----------------------------+---------+-----------------+
|             id            |            int             |         | attr_test_db.a1 |
|       int_array_col       |         array<int>         |         |                 |
|      array_array_col      |     array<array<int>>      |         |                 |
|       map_array_col       |   array<map<string,int>>   |         |                 |
|      struct_array_col     |       array<struct<        |         |                 |
|                           |          f1:bigint,        |         |                 |
|                           |          f2:string         |         |                 |
|                           |             >>             |         |                 |
|        int_map_col        |      map<string,int>       |         |                 |
|       array_map_col       |   map<string,array<int>>   |         |                 |
|       struct_map_col      |     map<string,struct<     |         |                 |
|                           |          f1:bigint,        |         |                 |
|                           |          f2:string         |         |                 |
|                           |             >>             |         |                 |
|       int_struct_col      |          struct<           |         |                 |
|                           |           f1:int,          |         |                 |
|                           |            f2:int          |         |                 |
|                           |             >              |         |                 |
|     complex_struct_col    |          struct<           |         |                 |
|                           |           f1:int,          |         |                 |
|                           |        f2:array<int>,      |         |                 |
|                           |      f3:map<string,int>    |         |                 |
|                           |             >              |         |                 |
|     nested_struct_col     |          struct<           |         |                 |
|                           |           f1:int,          |         | attr_test_db.a1 |
|                           |          f2:struct<        |         |                 |
|                           |          f11:bigint,       |         |                 |
|                           |          f12:struct<       |         |                 |
|                           |            f21:bigint      |         |                 |
|                           |               >            |         |                 |
|                           |              >             |         |                 |
|                           |             >              |         |                 |
| complex_nested_struct_col |          struct<           |         |                 |
|                           |           f1:int,          |         |                 |
|                           |       f2:array<struct<     |         |                 |
|                           |          f11:bigint,       |         |                 |
|                           |     f12:map<string,struct< |         |                 |
|                           |            f21:bigint      |         |                 |
|                           |               >>           |         |                 |
|                           |              >>            |         |                 |
|                           |             >              |         |                 |
|            item           |            int             |         |                 |
|            year           |            int             |         |                 |
|           month           |            int             |         |                 |
+---------------------------+----------------------------+---------+-----------------+""".strip())

            self._validate(conn, db, 'a1',
                [['functional', 'allcomplextypes', 'array_array_col'],
                 ['functional', 'allcomplextypes', 'nested_struct_col.f2.f12.f21']],
                """select * from functional.allcomplextypes,
                   functional.allcomplextypes.int_map_col""",
"""
+---------------------------+----------------------------+---------+-----------------+
|            name           |            type            | comment |    attributes   |
+---------------------------+----------------------------+---------+-----------------+
|             id            |            int             |         |                 |
|       int_array_col       |         array<int>         |         |                 |
|      array_array_col      |     array<array<int>>      |         | attr_test_db.a1 |
|       map_array_col       |   array<map<string,int>>   |         |                 |
|      struct_array_col     |       array<struct<        |         |                 |
|                           |          f1:bigint,        |         |                 |
|                           |          f2:string         |         |                 |
|                           |             >>             |         |                 |
|        int_map_col        |      map<string,int>       |         |                 |
|       array_map_col       |   map<string,array<int>>   |         |                 |
|       struct_map_col      |     map<string,struct<     |         |                 |
|                           |          f1:bigint,        |         |                 |
|                           |          f2:string         |         |                 |
|                           |             >>             |         |                 |
|       int_struct_col      |          struct<           |         |                 |
|                           |           f1:int,          |         |                 |
|                           |            f2:int          |         |                 |
|                           |             >              |         |                 |
|     complex_struct_col    |          struct<           |         |                 |
|                           |           f1:int,          |         |                 |
|                           |        f2:array<int>,      |         |                 |
|                           |      f3:map<string,int>    |         |                 |
|                           |             >              |         |                 |
|     nested_struct_col     |          struct<           |         |                 |
|                           |           f1:int,          |         |                 |
|                           |          f2:struct<        |         |                 |
|                           |          f11:bigint,       |         |                 |
|                           |          f12:struct<       |         |                 |
|                           |            f21:bigint      |         | attr_test_db.a1 |
|                           |               >            |         |                 |
|                           |              >             |         |                 |
|                           |             >              |         |                 |
| complex_nested_struct_col |          struct<           |         |                 |
|                           |           f1:int,          |         |                 |
|                           |       f2:array<struct<     |         |                 |
|                           |          f11:bigint,       |         |                 |
|                           |     f12:map<string,struct< |         |                 |
|                           |            f21:bigint      |         |                 |
|                           |               >>           |         |                 |
|                           |              >>            |         |                 |
|                           |             >              |         |                 |
|            key            |           string           |         |                 |
|           value           |            int             |         |                 |
|            year           |            int             |         |                 |
|           month           |            int             |         |                 |
+---------------------------+----------------------------+---------+-----------------+""".strip())

            #
            # TODO: maps are not supported
            #
            self._validate(conn, db, 'a1',
                [['functional', 'allcomplextypes', 'complex_nested_struct_col.f2.f12.key']],
                """select f12.key from functional.allcomplextypes,
                   functional.allcomplextypes.complex_nested_struct_col.f2.f12""",
"""
+------+--------+---------+------------+
| name |  type  | comment | attributes |
+------+--------+---------+------------+
| key  | string |         |            |
+------+--------+---------+------------+""".strip())

            self._validate(conn, db, 'a1',
                [['functional', 'allcomplextypes', 'int_array_col']],
                """select a.id, int_array_col.item from
                   functional.allcomplextypes a, a.int_array_col""",
"""
+------+------+---------+-----------------+
| name | type | comment |    attributes   |
+------+------+---------+-----------------+
|  id  | int  |         |                 |
| item | int  |         | attr_test_db.a1 |
+------+------+---------+-----------------+""".strip())

    def test_full_unnest_attributes(self):
        db = 'attr_test_db'
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            conn.execute_ddl("drop database if exists %s cascade" % db)
            conn.execute_ddl("create database if not exists %s" % db)

            # No tags
            self._validate(conn, db, 'a1', [],
                'select ** from chase.zd1238_4',
                """
+-----------------------------------------------+---------+---------+------------+
|                      name                     |   type  | comment | attributes |
+-----------------------------------------------+---------+---------+------------+
|                subscriptionkey                |  string |         |            |
|              partyroles__partykey             |  string |         |            |
|             partyroles__tenantkey             |  string |         |            |
|                partyroles__role               |  string |         |            |
|            partyroles__partyrolekey           |  string |         |            |
|            partyroles__createddate            |  string |         |            |
|            partyroles__updateddate            |  string |         |            |
|                   productkey                  |  string |         |            |
|                 productversion                |   int   |         |            |
|                  producttype                  |  string |         |            |
|             parentsubscriptionkey             |  string |         |            |
|              parentaccountnumber              |  string |         |            |
|                 accountnumber                 |  string |         |            |
|                    sortcode                   |  string |         |            |
|                  productname                  |  string |         |            |
|               requiredexternalid              |  string |         |            |
|               currency__country               |  string |         |            |
|             currency__currencycode            |  string |         |            |
|                 periodiccycle                 |  string |         |            |
|               subscriptionstatus              |  string |         |            |
|                  createddate                  |  string |         |            |
|                  updateddate                  |  string |         |            |
|    feeconfigurations__applicationfrequency    |  string |         |            |
|   feeconfigurations__chargeincludedindicator  | boolean |         |            |
|        feeconfigurations__chargingcycle       |  string |         |            |
|     feeconfigurations__decision__tablekey     |  string |         |            |
|       feeconfigurations__decision__input      |  string |         |            |
| feeconfigurations__event__eventidentification |  string |         |            |
|      feeconfigurations__event__eventname      |  string |         |            |
|          feeconfigurations__feeamount         |  string |         |            |
|      feeconfigurations__feecap__capamount     |  string |         |            |
|    feeconfigurations__feecap__capoccurrence   |  string |         |            |
|    feeconfigurations__feecap__cappingperiod   |  string |         |            |
|         feeconfigurations__feecategory        |  string |         |            |
|      feeconfigurations__feeidentification     |  string |         |            |
|           feeconfigurations__feename          |  string |         |            |
|        feeconfigurations__feerate__rate       |  string |         |            |
|   feeconfigurations__feerate__notionalvalue   |  string |         |            |
|           feeconfigurations__feetype          |  string |         |            |
|         feeconfigurations__maximumfee         |  string |         |            |
|         feeconfigurations__minimumfee         |  string |         |            |
|           feeconfigurations__taxrate          |  string |         |            |
|    feeconfigurations__statementdescription    |  string |         |            |
|   feeconfigurations__validityperiod__enddate  |  string |         |            |
|  feeconfigurations__validityperiod__startdate |  string |         |            |
|            kafka_message_timestamp            |  bigint |         |            |
+-----------------------------------------------+---------+---------+------------+""".strip())

            self._validate(conn, db, 'a1',
                [['chase', 'zd1238_4', 'productkey'],
                 ['chase', 'zd1238_4', 'partyroles.partykey']],
                'select ** from chase.zd1238_4',
                """
+-----------------------------------------------+---------+---------+-----------------+
|                      name                     |   type  | comment |    attributes   |
+-----------------------------------------------+---------+---------+-----------------+
|                subscriptionkey                |  string |         |                 |
|              partyroles__partykey             |  string |         | attr_test_db.a1 |
|             partyroles__tenantkey             |  string |         |                 |
|                partyroles__role               |  string |         |                 |
|            partyroles__partyrolekey           |  string |         |                 |
|            partyroles__createddate            |  string |         |                 |
|            partyroles__updateddate            |  string |         |                 |
|                   productkey                  |  string |         | attr_test_db.a1 |
|                 productversion                |   int   |         |                 |
|                  producttype                  |  string |         |                 |
|             parentsubscriptionkey             |  string |         |                 |
|              parentaccountnumber              |  string |         |                 |
|                 accountnumber                 |  string |         |                 |
|                    sortcode                   |  string |         |                 |
|                  productname                  |  string |         |                 |
|               requiredexternalid              |  string |         |                 |
|               currency__country               |  string |         |                 |
|             currency__currencycode            |  string |         |                 |
|                 periodiccycle                 |  string |         |                 |
|               subscriptionstatus              |  string |         |                 |
|                  createddate                  |  string |         |                 |
|                  updateddate                  |  string |         |                 |
|    feeconfigurations__applicationfrequency    |  string |         |                 |
|   feeconfigurations__chargeincludedindicator  | boolean |         |                 |
|        feeconfigurations__chargingcycle       |  string |         |                 |
|     feeconfigurations__decision__tablekey     |  string |         |                 |
|       feeconfigurations__decision__input      |  string |         |                 |
| feeconfigurations__event__eventidentification |  string |         |                 |
|      feeconfigurations__event__eventname      |  string |         |                 |
|          feeconfigurations__feeamount         |  string |         |                 |
|      feeconfigurations__feecap__capamount     |  string |         |                 |
|    feeconfigurations__feecap__capoccurrence   |  string |         |                 |
|    feeconfigurations__feecap__cappingperiod   |  string |         |                 |
|         feeconfigurations__feecategory        |  string |         |                 |
|      feeconfigurations__feeidentification     |  string |         |                 |
|           feeconfigurations__feename          |  string |         |                 |
|        feeconfigurations__feerate__rate       |  string |         |                 |
|   feeconfigurations__feerate__notionalvalue   |  string |         |                 |
|           feeconfigurations__feetype          |  string |         |                 |
|         feeconfigurations__maximumfee         |  string |         |                 |
|         feeconfigurations__minimumfee         |  string |         |                 |
|           feeconfigurations__taxrate          |  string |         |                 |
|    feeconfigurations__statementdescription    |  string |         |                 |
|   feeconfigurations__validityperiod__enddate  |  string |         |                 |
|  feeconfigurations__validityperiod__startdate |  string |         |                 |
|            kafka_message_timestamp            |  bigint |         |                 |
+-----------------------------------------------+---------+---------+-----------------+""".strip())

            self._validate(conn, db, 'a1',
                [['rs_complex', 'strarray_t', 'str_arr']],
                'select ** from rs_complex.strarray_t',
                """
+---------+--------+---------+-----------------+
|   name  |  type  | comment |    attributes   |
+---------+--------+---------+-----------------+
|    id   | bigint |         |                 |
| str_arr | string |         | attr_test_db.a1 |
+---------+--------+---------+-----------------+""".strip())

            self._validate(conn, db, 'a1',
                [['rs_complex', 'array_struct_t', 'a1.f1']],
                'select ** from rs_complex.array_struct_t',
                """
+--------+--------+---------+-----------------+
|  name  |  type  | comment |    attributes   |
+--------+--------+---------+-----------------+
| a1__f1 | string |         | attr_test_db.a1 |
| a1__f2 | string |         |                 |
+--------+--------+---------+-----------------+""".strip())

            self._validate(conn, db, 'a1',
                [['rs_complex', 'array_struct_t', 'a1.f1'],
                 ['rs_complex', 'array_struct_t', 'a1.f2']],
                'select ** from rs_complex.array_struct_t',
                """
+--------+--------+---------+-----------------+
|  name  |  type  | comment |    attributes   |
+--------+--------+---------+-----------------+
| a1__f1 | string |         | attr_test_db.a1 |
| a1__f2 | string |         | attr_test_db.a1 |
+--------+--------+---------+-----------------+""".strip())

            self._validate(conn, db, 'a1',
                [['rs_complex', 'array_struct_array', 'a1.a2'],
                 ['rs_complex', 'array_struct_array', 'a1.f1'],
                 ['rs_complex', 'array_struct_array', 'a1.f2']],
                'select ** from rs_complex.array_struct_array',
                """
+--------+--------+---------+-----------------+
|  name  |  type  | comment |    attributes   |
+--------+--------+---------+-----------------+
| a1__f1 | string |         | attr_test_db.a1 |
| a1__f2 | string |         | attr_test_db.a1 |
| a1__a2 | string |         | attr_test_db.a1 |
+--------+--------+---------+-----------------+""".strip())

# FIXME: map is not supported.
#            self._validate(conn, db, 'a1',
#                [['rs_complex_parquet', 'spark_gzip', 'int32'],
#                 ['rs_complex_parquet', 'spark_gzip', 'str_arr'],
#                 ['rs_complex_parquet', 'spark_gzip', 'int_arr']],
#                'select ** from rs_complex_parquet.spark_gzip',
#                """
#+---------+--------+---------+-----------------+
#|   name  |  type  | comment |    attributes   |
#+---------+--------+---------+-----------------+
#|  int32  |  int   |         | attr_test_db.a1 |
#| str_arr | string |         | attr_test_db.a1 |
#| int_arr |  int   |         | attr_test_db.a1 |
#+---------+--------+---------+-----------------+""".strip())

            self._validate(conn, db, 'a1',
                [['chase', 'zd1211', 'accountnumber'],
                 ['chase', 'zd1211', 'productdetails.subproducts']],
                'select ** from chase.zd1211',
"""
+---------------------------------------------------------------------------------------------------------+---------+---------+-----------------+
|                                                   name                                                  |   type  | comment |    attributes   |
+---------------------------------------------------------------------------------------------------------+---------+---------+-----------------+
|                                             subscriptionkey                                             |  string |         |                 |
|                                           partyroles__partykey                                          |  string |         |                 |
|                                          partyroles__tenantkey                                          |  string |         |                 |
|                                             partyroles__role                                            |  string |         |                 |
|                                         partyroles__partyrolekey                                        |  string |         |                 |
|                                         partyroles__createddate                                         |  string |         |                 |
|                                         partyroles__updateddate                                         |  string |         |                 |
|                                               productname                                               |  string |         |                 |
|                                              effectivedate                                              |  string |         |                 |
|                                                productkey                                               |  string |         |                 |
|                                              productversion                                             |   int   |         |                 |
|                                          parentsubscriptionkey                                          |  string |         |                 |
|                                           parentaccountnumber                                           |  string |         |                 |
|                                              accountnumber                                              |  string |         | attr_test_db.a1 |
|                                                 sortcode                                                |  string |         |                 |
|                                            subscriptionstatus                                           |  string |         |                 |
|                                               createddate                                               |  string |         |                 |
|                                               updateddate                                               |  string |         |                 |
|                                          linkedsubscriptionkey                                          |  string |         |                 |
|                                      productdetails__effectivedate                                      |  string |         |                 |
|                                        productdetails__productkey                                       |  string |         |                 |
|                                        productdetails__tenantkey                                        |  string |         |                 |
|                                     productdetails__productcategory                                     |  string |         |                 |
|                                       productdetails__productname                                       |  string |         |                 |
|                                       productdetails__producttype                                       |  string |         |                 |
|                                    productdetails__productdescription                                   |  string |         |                 |
|                                      productdetails__productsegment                                     |  string |         |                 |
|                                       productdetails__createddate                                       |  string |         |                 |
|                                        productdetails__createdby                                        |  string |         |                 |
|                                       productdetails__updateddate                                       |  string |         |                 |
|                                        productdetails__updatedby                                        |  string |         |                 |
|                                      productdetails__publisheddate                                      |  string |         |                 |
|                                       productdetails__publishedby                                       |  string |         |                 |
|                                        productdetails__closeddate                                       |  string |         |                 |
|                                         productdetails__closedby                                        |  string |         |                 |
|                                          productdetails__status                                         |  string |         |                 |
|                                       productdetails__majorversion                                      |   int   |         |                 |
|                                           productdetails__tags                                          |  string |         |                 |
|                                    productdetails__linkedproduct__id                                    |  string |         |                 |
|                                       productdetails__subproducts                                       |  string |         | attr_test_db.a1 |
|                                    productdetails__requiredexternalid                                   |  string |         |                 |
|                        productdetails__limits__transactionlimits__transactionname                       |  string |         |                 |
|                          productdetails__limits__transactionlimits__description                         |  string |         |                 |
|                         productdetails__limits__transactionlimits__minimumamount                        |  string |         |                 |
|                         productdetails__limits__transactionlimits__maximumamount                        |  string |         |                 |
|                          productdetails__limits__transactionlimits__resetperiod                         |  string |         |                 |
|                             productdetails__limits__schemelimits__schemename                            |  string |         |                 |
|                            productdetails__limits__schemelimits__description                            |  string |         |                 |
|                           productdetails__limits__schemelimits__minimumamount                           |  string |         |                 |
|                           productdetails__limits__schemelimits__maximumamount                           |  string |         |                 |
|                            productdetails__limits__schemelimits__resetperiod                            |  string |         |                 |
|                        productdetails__limits__accountbalancelimits__balancetype                        |  string |         |                 |
|                        productdetails__limits__accountbalancelimits__description                        |  string |         |                 |
|                       productdetails__limits__accountbalancelimits__minimumamount                       |  string |         |                 |
|                       productdetails__limits__accountbalancelimits__maximumamount                       |  string |         |                 |
|                            productdetails__limits__productlimits__productkey                            |  string |         |                 |
|                            productdetails__limits__productlimits__productname                           |  string |         |                 |
|                            productdetails__limits__productlimits__producttype                           |  string |         |                 |
|                            productdetails__limits__productlimits__description                           |  string |         |                 |
|                           productdetails__limits__productlimits__maximumnumber                          |   int   |         |                 |
|                         productdetails__limits__fundinglimits__fundingmechanism                         |  string |         |                 |
|                            productdetails__limits__fundinglimits__description                           |  string |         |                 |
|                           productdetails__limits__fundinglimits__minimumamount                          |  string |         |                 |
|                           productdetails__limits__fundinglimits__maximumamount                          |  string |         |                 |
|                           productdetails__limits__fundinglimits__defaultamount                          |  string |         |                 |
|                                 productdetails__cards__cardstatusdefault                                |  string |         |                 |
|                                       productdetails__cards__type                                       |  string |         |                 |
|                                      productdetails__cards__scheme                                      |  string |         |                 |
|                        productdetails__cards__channeldefaultsettings__contactless                       | boolean |         |                 |
|                          productdetails__cards__channeldefaultsettings__chippin                         | boolean |         |                 |
|                            productdetails__cards__channeldefaultsettings__atm                           | boolean |         |                 |
|                            productdetails__cards__channeldefaultsettings__cnp                           | boolean |         |                 |
|                         productdetails__cards__channeldefaultsettings__magstripe                        | boolean |         |                 |
|                       productdetails__cards__channeldefaultsettings__international                      | boolean |         |                 |
|                          productdetails__cards__channeldefaultsettings__online                          | boolean |         |                 |
|                              productdetails__termsandconditions__identifier                             |  string |         |                 |
|                                 productdetails__termsandconditions__name                                |  string |         |                 |
|                             productdetails__termsandconditions__description                             |  string |         |                 |
|                               productdetails__termsandconditions__version                               |  string |         |                 |
|                             productdetails__termsandconditions__createddate                             |  string |         |                 |
|                              productdetails__termsandconditions__createdby                              |  string |         |                 |
|                             productdetails__termsandconditions__updateddate                             |  string |         |                 |
|                              productdetails__termsandconditions__updatedby                              |  string |         |                 |
|                            productdetails__termsandconditions__publisheddate                            |  string |         |                 |
|                                productdetails__termsandconditions__status                               |  string |         |                 |
|                              productdetails__termsandconditions__files__url                             |  string |         |                 |
|                            productdetails__termsandconditions__files__format                            |  string |         |                 |
|                        productdetails__internaldocumentation__salesaccesschannels                       |  string |         |                 |
|                       productdetails__internaldocumentation__servingaccesschannels                      |  string |         |                 |
|                           productdetails__internaldocumentation__mobilewallet                           |  string |         |                 |
|                            productdetails__internaldocumentation__producturl                            |  string |         |                 |
|                          productdetails__internaldocumentation__fraudriskrating                         |  string |         |                 |
|                        productdetails__internaldocumentation__fincrimeriskrating                        |  string |         |                 |
|                          productdetails__subscriptioncreationrule__lockedstatus                         | boolean |         |                 |
|                   productdetails__subscriptioncreationrule__requiredsignatoriesnumber                   |   int   |         |                 |
|                   productdetails__subscriptioncreationrule__defaultsubscriptionstatus                   |  string |         |                 |
|                                  productdetails__currency__currencycode                                 |  string |         |                 |
|                                    productdetails__currency__country                                    |  string |         |                 |
|                                      productdetails__teams__teamkey                                     |  string |         |                 |
|                                       productdetails__teams__write                                      | boolean |         |                 |
|                                  productdetails__customattributes__name                                 |  string |         |                 |
|                                 productdetails__customattributes__value                                 |  string |         |                 |
|                                  productdetails__documents__identifier                                  |  string |         |                 |
|                                     productdetails__documents__name                                     |  string |         |                 |
|                                  productdetails__documents__description                                 |  string |         |                 |
|                                    productdetails__documents__version                                   |  string |         |                 |
|                                  productdetails__documents__createddate                                 |  string |         |                 |
|                                   productdetails__documents__createdby                                  |  string |         |                 |
|                                  productdetails__documents__updateddate                                 |  string |         |                 |
|                                   productdetails__documents__updatedby                                  |  string |         |                 |
|                                 productdetails__documents__publisheddate                                |  string |         |                 |
|                                    productdetails__documents__status                                    |  string |         |                 |
|                                  productdetails__documents__files__url                                  |  string |         |                 |
|                                 productdetails__documents__files__format                                |  string |         |                 |
|                                 productdetails__feescharges__feecategory                                |  string |         |                 |
|                                   productdetails__feescharges__feetype                                  |  string |         |                 |
|                               productdetails__feescharges__feedescription                               |  string |         |                 |
|                             productdetails__feescharges__feetransactioncode                             |  string |         |                 |
|                               productdetails__feescharges__transactioncode                              |  string |         |                 |
|                                   productdetails__feescharges__taxrate                                  |  string |         |                 |
|                            productdetails__feescharges__statementdescription                            |  string |         |                 |
|                                   productdetails__feescharges__feename                                  |  string |         |                 |
|                                  productdetails__feescharges__feeamount                                 |  string |         |                 |
|                           productdetails__feescharges__taxstatementdescription                          |  string |         |                 |
|                            productdetails__feescharges__calculationfrequency                            |  string |         |                 |
|                            productdetails__feescharges__applicationfrequency                            |  string |         |                 |
|                                productdetails__feescharges__feerate__rate                               |  string |         |                 |
|                           productdetails__feescharges__feerate__notionalvalue                           |  string |         |                 |
|                          productdetails__feescharges__feerate__maximumfeeamount                         |  string |         |                 |
|                          productdetails__feescharges__feerate__minimumfeeamount                         |  string |         |                 |
|                          productdetails__feescharges__validityperiod__startdate                         |  string |         |                 |
|                           productdetails__feescharges__validityperiod__enddate                          |  string |         |                 |
|                           productdetails__feescharges__validityperiod__period                           |  string |         |                 |
|                        productdetails__feescharges__validityperiod__periodamount                        |  string |         |                 |
|                            productdetails__feescharges__feecap__cappingperiod                           |  string |         |                 |
|                            productdetails__feescharges__feecap__feecapamount                            |  string |         |                 |
|                          productdetails__feescharges__feecap__feecapoccurrence                          |  string |         |                 |
|                       productdetails__feescharges__feerules__inputparameters__name                      |  string |         |                 |
|                      productdetails__feescharges__feerules__inputparameters__value                      |  string |         |                 |
|                     productdetails__feescharges__feerules__inputparameters__operator                    |  string |         |                 |
|                      productdetails__feescharges__feerules__outputparameters__name                      |  string |         |                 |
|                      productdetails__feescharges__feerules__outputparameters__value                     |  string |         |                 |
|                           productdetails__feescharges__notificationapplication                          |   int   |         |                 |
|                            productdetails__creditinterest__fixedvariabletype                            |  string |         |                 |
|                               productdetails__creditinterest__includefees                               | boolean |         |                 |
|                              productdetails__creditinterest__roundingmethod                             |  string |         |                 |
|                                 productdetails__creditinterest__daycount                                |  string |         |                 |
|                               productdetails__creditinterest__interestrate                              |  string |         |                 |
|                            productdetails__creditinterest__tierbandcalcmethod                           |  string |         |                 |
|                           productdetails__creditinterest__applicationfrequency                          |  string |         |                 |
|                           productdetails__creditinterest__compoundingfrequency                          |  string |         |                 |
|                           productdetails__creditinterest__calculationfrequency                          |  string |         |                 |
|                             productdetails__creditinterest__balancecriteria                             |  string |         |                 |
|                       productdetails__creditinterest__interesttierband__startrange                      |  string |         |                 |
|                        productdetails__creditinterest__interesttierband__endrange                       |  string |         |                 |
|                      productdetails__creditinterest__interesttierband__tierbandrate                     |  string |         |                 |
|                         productdetails__creditinterest__interestrateindex__index                        |  string |         |                 |
|                         productdetails__creditinterest__interestrateindex__term                         |  string |         |                 |
|                      productdetails__creditinterest__interestrateindex__identifier                      |  string |         |                 |
|                          productdetails__creditinterest__interestrateindex__url                         |  string |         |                 |
|                             productdetails__overdraft__applicationfrequency                             |  string |         |                 |
|                                productdetails__overdraft__balancecriteria                               |  string |         |                 |
|                             productdetails__overdraft__calculationfrequency                             |  string |         |                 |
|                             productdetails__overdraft__compoundingfrequency                             |  string |         |                 |
|                                   productdetails__overdraft__daycount                                   |  string |         |                 |
|                                 productdetails__overdraft__interestrate                                 |  string |         |                 |
|                           productdetails__overdraft__overdrafttierband__buffer                          |  string |         |                 |
|                          productdetails__overdraft__overdrafttierband__endrange                         |  string |         |                 |
|                        productdetails__overdraft__overdrafttierband__interestrate                       |  string |         |                 |
|         productdetails__overdraft__overdrafttierband__overdraftfeescharges__applicationfrequency        |  string |         |                 |
|         productdetails__overdraft__overdrafttierband__overdraftfeescharges__calculationfrequency        |  string |         |                 |
|              productdetails__overdraft__overdrafttierband__overdraftfeescharges__feeamount              |  string |         |                 |
|        productdetails__overdraft__overdrafttierband__overdraftfeescharges__feecap__cappingperiod        |  string |         |                 |
|         productdetails__overdraft__overdrafttierband__overdraftfeescharges__feecap__feecapamount        |  string |         |                 |
|       productdetails__overdraft__overdrafttierband__overdraftfeescharges__feecap__feecapoccurrence      |  string |         |                 |
|             productdetails__overdraft__overdrafttierband__overdraftfeescharges__feecategory             |  string |         |                 |
|            productdetails__overdraft__overdrafttierband__overdraftfeescharges__feedescription           |  string |         |                 |
|   productdetails__overdraft__overdrafttierband__overdraftfeescharges__feerules__inputparameters__name   |  string |         |                 |
|   productdetails__overdraft__overdrafttierband__overdraftfeescharges__feerules__inputparameters__value  |  string |         |                 |
| productdetails__overdraft__overdrafttierband__overdraftfeescharges__feerules__inputparameters__operator |  string |         |                 |
|   productdetails__overdraft__overdrafttierband__overdraftfeescharges__feerules__outputparameters__name  |  string |         |                 |
|  productdetails__overdraft__overdrafttierband__overdraftfeescharges__feerules__outputparameters__value  |  string |         |                 |
|          productdetails__overdraft__overdrafttierband__overdraftfeescharges__feetransactioncode         |  string |         |                 |
|         productdetails__overdraft__overdrafttierband__overdraftfeescharges__statementdescription        |  string |         |                 |
|               productdetails__overdraft__overdrafttierband__overdraftfeescharges__taxrate               |  string |         |                 |
|      productdetails__overdraft__overdrafttierband__overdraftfeescharges__validityperiod__startdate      |  string |         |                 |
|       productdetails__overdraft__overdrafttierband__overdraftfeescharges__validityperiod__enddate       |  string |         |                 |
|        productdetails__overdraft__overdrafttierband__overdraftfeescharges__validityperiod__period       |  string |         |                 |
|     productdetails__overdraft__overdrafttierband__overdraftfeescharges__validityperiod__periodamount    |  string |         |                 |
|       productdetails__overdraft__overdrafttierband__overdraftfeescharges__notificationapplication       |   int   |         |                 |
|                         productdetails__overdraft__overdrafttierband__startrange                        |  string |         |                 |
|                          productdetails__overdraft__overdrafttierband__tiername                         |  string |         |                 |
|                     productdetails__overdraft__overdrafttierband__tierstatusdefault                     |  string |         |                 |
|                                productdetails__overdraft__roundingmethod                                |  string |         |                 |
|                        productdetails__overdraft__termsandconditions__identifier                        |  string |         |                 |
|                           productdetails__overdraft__termsandconditions__name                           |  string |         |                 |
|                        productdetails__overdraft__termsandconditions__description                       |  string |         |                 |
|                          productdetails__overdraft__termsandconditions__version                         |  string |         |                 |
|                        productdetails__overdraft__termsandconditions__createddate                       |  string |         |                 |
|                         productdetails__overdraft__termsandconditions__createdby                        |  string |         |                 |
|                        productdetails__overdraft__termsandconditions__updateddate                       |  string |         |                 |
|                         productdetails__overdraft__termsandconditions__updatedby                        |  string |         |                 |
|                       productdetails__overdraft__termsandconditions__publisheddate                      |  string |         |                 |
|                          productdetails__overdraft__termsandconditions__status                          |  string |         |                 |
|                        productdetails__overdraft__termsandconditions__files__url                        |  string |         |                 |
|                       productdetails__overdraft__termsandconditions__files__format                      |  string |         |                 |
|                              productdetails__overdraft__tierbandcalcmethod                              |  string |         |                 |
|                            productdetails__overdraft__notificationapplication                           |   int   |         |                 |
|                         productdetails__eligibility__ageeligibility__maximumage                         |   int   |         |                 |
|                         productdetails__eligibility__ageeligibility__minimumage                         |   int   |         |                 |
|                productdetails__eligibility__creditcheckeligibility__maximumpersonalincome               |  string |         |                 |
|                    productdetails__eligibility__creditcheckeligibility__maximumscore                    |  string |         |                 |
|                productdetails__eligibility__creditcheckeligibility__minimumpersonalincome               |  string |         |                 |
|                    productdetails__eligibility__creditcheckeligibility__minimumscore                    |  string |         |                 |
|                    productdetails__eligibility__creditcheckeligibility__scoringmodel                    |  string |         |                 |
|                   productdetails__eligibility__creditcheckeligibility__scoringprovider                  |  string |         |                 |
|                   productdetails__eligibility__creditcheckeligibility__scoringsegment                   |  string |         |                 |
|                     productdetails__eligibility__creditcheckeligibility__scoringtype                    |  string |         |                 |
|                           productdetails__eligibility__ideligibility__idproof                           |  string |         |                 |
|                            productdetails__eligibility__ideligibility__idtype                           |  string |         |                 |
|                             productdetails__eligibility__ideligibility__url                             |  string |         |                 |
|                   productdetails__eligibility__industryeligibility__customindustryname                  |  string |         |                 |
|                        productdetails__eligibility__industryeligibility__siccode                        |  string |         |                 |
|                 productdetails__eligibility__legalstructureeligibility__countryincluded                 |  string |         |                 |
|                  productdetails__eligibility__legalstructureeligibility__legalstructure                 |  string |         |                 |
|                  productdetails__eligibility__legalstructureeligibility__stateincluded                  |  string |         |                 |
|                        productdetails__eligibility__officereligibility__maxamount                       |  string |         |                 |
|                        productdetails__eligibility__officereligibility__minamount                       |  string |         |                 |
|                       productdetails__eligibility__officereligibility__officertype                      |  string |         |                 |
|                          productdetails__eligibility__othereligibility__amount                          |  string |         |                 |
|                        productdetails__eligibility__othereligibility__description                       |  string |         |                 |
|                         productdetails__eligibility__othereligibility__indicator                        | boolean |         |                 |
|                           productdetails__eligibility__othereligibility__name                           |  string |         |                 |
|                          productdetails__eligibility__othereligibility__period                          |  string |         |                 |
|                           productdetails__eligibility__othereligibility__type                           |  string |         |                 |
|                 productdetails__eligibility__professioneligibility__customprofessionname                |  string |         |                 |
|                       productdetails__eligibility__professioneligibility__soccode                       |  string |         |                 |
|                    productdetails__eligibility__residencyeligibility__countryincluded                   |  string |         |                 |
|                     productdetails__eligibility__residencyeligibility__minimumperiod                    |  string |         |                 |
|                  productdetails__eligibility__residencyeligibility__minimumperiodamount                 |  string |         |                 |
|                    productdetails__eligibility__residencyeligibility__residencestatus                   |  string |         |                 |
|                     productdetails__eligibility__residencyeligibility__residencytype                    |  string |         |                 |
|                     productdetails__eligibility__residencyeligibility__stateincluded                    |  string |         |                 |
|                      productdetails__eligibility__tradinghistoryeligibility__amount                     |  string |         |                 |
|                   productdetails__eligibility__tradinghistoryeligibility__description                   |  string |         |                 |
|                    productdetails__eligibility__tradinghistoryeligibility__indicator                    | boolean |         |                 |
|                    productdetails__eligibility__tradinghistoryeligibility__minmaxtype                   |  string |         |                 |
|                      productdetails__eligibility__tradinghistoryeligibility__period                     |  string |         |                 |
|                   productdetails__eligibility__tradinghistoryeligibility__tradingtype                   |  string |         |                 |
|                                 productdetails__statement__statementtype                                |  string |         |                 |
|                                productdetails__statement__statementperiod                               |  string |         |                 |
|                             productdetails__statement__statementdescription                             |  string |         |                 |
|                                   productdetails__statement__startday                                   |   int   |         |                 |
|                                  productdetails__statement__startmonth                                  |   int   |         |                 |
|                                productdetails__statements__statementtype                                |  string |         |                 |
|                               productdetails__statements__statementperiod                               |  string |         |                 |
|                             productdetails__statements__statementdescription                            |  string |         |                 |
|                                   productdetails__statements__startday                                  |   int   |         |                 |
|                                  productdetails__statements__startmonth                                 |   int   |         |                 |
|                                         kafka_message_timestamp                                         |  bigint |         |                 |
+---------------------------------------------------------------------------------------------------------+---------+---------+-----------------+""".strip())

            self._validate(conn, db, 'a1',
                [['chase', 'party', 'document.imagetype']],
                'select ** from chase.party',
"""
+----------------------------+-----------+---------+-----------------+
|            name            |    type   | comment |    attributes   |
+----------------------------+-----------+---------+-----------------+
|          partykey          |   string  |         |                 |
|         tenantkey          |   string  |         |                 |
|         givenname          |   string  |         |                 |
|          lastname          |   string  |         |                 |
|         middlename         |   string  |         |                 |
|       preferredname        |   string  |         |                 |
|        mobilenumber        |   string  |         |                 |
|           email            |   string  |         |                 |
|   address__addressline1    |   string  |         |                 |
|   address__addressline2    |   string  |         |                 |
|   address__addressline3    |   string  |         |                 |
|   address__addressline4    |   string  |         |                 |
|   address__addressline5    |   string  |         |                 |
|     address__postcode      |   string  |         |                 |
|       address__city        |   string  |         |                 |
|      address__country      |   string  |         |                 |
|       address__state       |   string  |         |                 |
|    address__addresstype    |   string  |         |                 |
|      address__status       |   string  |         |                 |
|    address__createddate    |   string  |         |                 |
|    address__updateddate    |   string  |         |                 |
|         birthdate          |    int    |         |                 |
|           status           |   string  |         |                 |
|     devices__deviceid      |   string  |         |                 |
|     devices__pushtoken     |   string  |         |                 |
|    devices__devicetype     |   string  |         |                 |
|    devices__createddate    |   string  |         |                 |
|    devices__updateddate    |   string  |         |                 |
|  ecis__externalidentifier  |   string  |         |                 |
|       ecis__provider       |   string  |         |                 |
|     ecis__createddate      |   string  |         |                 |
|     ecis__updateddate      |   string  |         |                 |
|        document__id        |   string  |         |                 |
|     document__partykey     |   string  |         |                 |
|    document__imagetype     |   string  |         | attr_test_db.a1 |
|  document__documentnumber  |   string  |         |                 |
|  document__issuingcountry  |   string  |         |                 |
|   document__dateofexpiry   |   string  |         |                 |
|   document__dateofbirth    |   string  |         |                 |
|     document__surname      |   string  |         |                 |
|    document__firstname     |   string  |         |                 |
|   document__placeofbirth   |   string  |         |                 |
|   document__createddate    |   string  |         |                 |
|   document__updateddate    |   string  |         |                 |
|        createddate         |   string  |         |                 |
|        updateddate         |   string  |         |                 |
|   citizenship__partykey    |   string  |         |                 |
| citizenship__nationalities |   string  |         |                 |
|  citizenship__createddate  |   string  |         |                 |
|  citizenship__updateddate  |   string  |         |                 |
|  kafka_message_timestamp   | timestamp |         |                 |
+----------------------------+-----------+---------+-----------------+""".strip())

            self._validate(conn, db, 'a1',
                [['rs_complex', 'array_struct_array', 'a1.f1'],
                 ['rs_complex', 'array_struct_array', 'a1.a2']],
                'select ** from rs_complex.array_struct_array',
"""
+--------+--------+---------+-----------------+
|  name  |  type  | comment |    attributes   |
+--------+--------+---------+-----------------+
| a1__f1 | string |         | attr_test_db.a1 |
| a1__f2 | string |         |                 |
| a1__a2 | string |         | attr_test_db.a1 |
+--------+--------+---------+-----------------+""".strip())

# TODO: map support
#            self._validate(conn, db, 'a1',
#                [['functional', 'allcomplextypes', 'id'],
#                 ['functional', 'allcomplextypes', 'nested_struct_col.f1']],
#                'select ** from functional.allcomplextypes',
#"""
#+---------------------------+----------------------------+---------+-----------------+
#|            name           |            type            | comment |    attributes   |
#+---------------------------+----------------------------+---------+-----------------+
#|             id            |            int             |         | attr_test_db.a1 |
#|       int_array_col       |         array<int>         |         |                 |
#|      array_array_col      |     array<array<int>>      |         |                 |
#|       map_array_col       |   array<map<string,int>>   |         |                 |
#|      struct_array_col     |       array<struct<        |         |                 |
#|                           |          f1:bigint,        |         |                 |
#|                           |          f2:string         |         |                 |
#|                           |             >>             |         |                 |
#|        int_map_col        |      map<string,int>       |         |                 |
#|       array_map_col       |   map<string,array<int>>   |         |                 |
#|       struct_map_col      |     map<string,struct<     |         |                 |
#|                           |          f1:bigint,        |         |                 |
#|                           |          f2:string         |         |                 |
#|                           |             >>             |         |                 |
#|       int_struct_col      |          struct<           |         |                 |
#|                           |           f1:int,          |         |                 |
#|                           |            f2:int          |         |                 |
#|                           |             >              |         |                 |
#|     complex_struct_col    |          struct<           |         |                 |
#|                           |           f1:int,          |         |                 |
#|                           |        f2:array<int>,      |         |                 |
#|                           |      f3:map<string,int>    |         |                 |
#|                           |             >              |         |                 |
#|     nested_struct_col     |          struct<           |         |                 |
#|                           |           f1:int,          |         | attr_test_db.a1 |
#|                           |          f2:struct<        |         |                 |
#|                           |          f11:bigint,       |         |                 |
#|                           |          f12:struct<       |         |                 |
#|                           |            f21:bigint      |         |                 |
#|                           |               >            |         |                 |
#|                           |              >             |         |                 |
#|                           |             >              |         |                 |
#| complex_nested_struct_col |          struct<           |         |                 |
#|                           |           f1:int,          |         |                 |
#|                           |       f2:array<struct<     |         |                 |
#|                           |          f11:bigint,       |         |                 |
#|                           |     f12:map<string,struct< |         |                 |
#|                           |            f21:bigint      |         |                 |
#|                           |               >>           |         |                 |
#|                           |              >>            |         |                 |
#|                           |             >              |         |                 |
#|            item           |            int             |         |                 |
#|            year           |            int             |         |                 |
#|           month           |            int             |         |                 |
#+---------------------------+----------------------------+---------+-----------------+""".strip())

if __name__ == "__main__":
    unittest.main()



# Copyright 2019 Okera Inc. All Rights Reserved.
#
# Some integration tests for user attributes
#
# pylint: disable=broad-except
# pylint: disable=global-statement
# pylint: disable=no-self-use
# pylint: disable=too-many-lines
# pylint: disable=too-many-public-methods
# pylint: disable=too-many-locals
# pylint: disable=bad-continuation
# pylint: disable=broad-except

from okera import context
from okera.tests import pycerebro_test_common as common

from okera._thrift_api import (
    TGetUserAttributesParams, TRecordServiceException)

from okera import _thrift_api

class UserAttributesTest(common.TestBase):
    def test_all_users(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            # Generate some users and fetch user attributes for them so we
            # can get the values when we fetch it all.
            users = [common.random_string(10), common.random_string(10)]
            for user in users:
                ctx.enable_token_auth(token_str=user)
                conn.scan_as_json("select user_attribute('whoami')", dialect='okera')

            ctx.disable_auth()

            print("Checking user attributes for:", users)

            params = TGetUserAttributesParams()
            params.all_users = True
            results = conn.service.client.GetUserAttributes(params)
            all_attributes = results.attributes
            assert len(all_attributes) >= len(users)
            for user in users:
                assert user in all_attributes
                assert 'whoami' in all_attributes[user]
                assert all_attributes[user]['whoami'].value == user
                assert all_attributes[user]['whoami'].source == 'static'

            # Ensure we can't access this as a non-admin user
            with self.assertRaises(TRecordServiceException) as ex_ctx:
                params = TGetUserAttributesParams()
                params.requesting_user = users[0]
                params.all_users = True
                conn.service.client.GetUserAttributes(params)
            self.assertTrue('Only admins can request user attributes for all users' in str(ex_ctx.exception))

    def test_specific_user(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            # Generate some users and fetch user attributes for them so we
            # can get the values when we fetch it all.
            user = common.random_string(10)
            ctx.enable_token_auth(token_str=user)
            conn.scan_as_json("select user_attribute('whoami')", dialect='okera')

            ctx.disable_auth()

            print("Checking user attributes for:", user)

            params = TGetUserAttributesParams()
            params.user = user
            results = conn.service.client.GetUserAttributes(params)
            all_attributes = results.attributes
            assert len(all_attributes) >= 1
            assert user in all_attributes
            assert 'whoami' in all_attributes[user]
            assert all_attributes[user]['whoami'].value == user
            assert all_attributes[user]['whoami'].source == 'static'

            # Ensure we can access it as the same user
            params = TGetUserAttributesParams()
            params.requesting_user = user
            params.user = user
            results = conn.service.client.GetUserAttributes(params)
            all_attributes = results.attributes
            assert len(all_attributes) >= 1
            assert user in all_attributes
            assert 'whoami' in all_attributes[user]
            assert all_attributes[user]['whoami'].value == user
            assert all_attributes[user]['whoami'].source == 'static'

            # Ensure we can't access this as a non-admin user
            with self.assertRaises(TRecordServiceException) as ex_ctx:
                params = TGetUserAttributesParams()
                params.requesting_user = common.random_string(10)
                params.user = user
                conn.service.client.GetUserAttributes(params)
            self.assertTrue('user attributes for a user other than themselves' in str(ex_ctx.exception))

    def test_current_user(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            # Generate some users and fetch user attributes for them so we
            # can get the values when we fetch it all.
            users = [common.random_string(10), common.random_string(10)]
            for user in users:
                ctx.enable_token_auth(token_str=user)
                conn.scan_as_json("select user_attribute('whoami')", dialect='okera')

            print("Checking user attributes for:", users)

            for user in users:
                ctx.enable_token_auth(token_str=user)
                params = TGetUserAttributesParams()
                params.requesting_user = user
                results = conn.service.client.GetUserAttributes(params)
                all_attributes = results.attributes
                assert len(all_attributes) == 1
                assert user in all_attributes
                assert 'whoami' in all_attributes[user]
                assert all_attributes[user]['whoami'].value == user
                assert all_attributes[user]['whoami'].source == 'static'


# Copyright 2019 Okera Inc. All Rights Reserved.
#
# Some integration tests against the crawler

# pylint: disable=broad-except,line-too-long,too-many-public-methods
import os
import time
import unittest
from collections import namedtuple
import pytest
from okera import _thrift_api
from okera.tests import pycerebro_test_common as common

TEST_USER = "testcrawleruser"
TEST_USER2 = "testcrawleruser2"
TEST_ROLE = "test_crawler_role"
TEST_ROLE2 = "test_crawler_role2"

CRAWLER_STATUS_KEY = 'okera.crawler.status'
CRAWLER_DATA_FILE_PATH_KEY = 'okera.crawler.data_file_path'

"""
This function determines which cloud providers are currently
supported in the running test cluster.  It is not properly wired
to detect platform support, so this needs to be tweaked manually.

To run this test suite with Azure support, start cdas with:
    start-cdas -load-test-data-minimal -autotagger -azure
To run this with AWS support, start cdas with:
    start-cdas -load-test-data-minimal -autotagger

Execute the UTs with:
    cd $OKERA_HOME/client/pycerebro
    export CRAWLER_TEST_MODES="S3,ADLS,ABFS,AUTOTAGGER"
    pytest okera/tests/test_crawler.py
"""
# TODO: We need cleanup handling here. We should attempt to delete previous crawlers
# before starting the next one. Related ticket: DAS-3831

MAX_WAIT_SEC = 60 if os.environ.get('CONTAINERIZED', 'false') == 'true' else 25

CRAWLER_TEST_MODES = os.environ.get('CRAWLER_TEST_MODES', 'S3')

FIELDS = ['provider', 'name', 'root_path', 'found_tables', 'timeout']
CrawlerTest = namedtuple("CrawlerTest", FIELDS)
Privilege = namedtuple("Privilege", ['level', 'can_select', 'can_drop', 'can_alter'])

PRIV_ALL = Privilege('ALL', True, True, True)
PRIV_SHOW = Privilege('SHOW', False, False, False)
PRIV_SELECT = Privilege('SELECT', True, False, False)
PRIV_DROP = Privilege('DROP', False, True, False)
PRIV_ALTER = Privilege('ALTER', False, False, True)

@pytest.mark.skip(reason="Currently broken and fix being tracked by DAS-5760")
class CrawlerIntegrationTest(unittest.TestCase):
    def _wait_for_crawler(self, conn, crawler_db, dataset=None, max_sec=MAX_WAIT_SEC):
        """ Waits for the crawler to find dataset for this crawler
            If dataset is None, then just wait for the crawler to report COMPLETED.
        """
        sleep_time = 1
        if max_sec > 60:
            sleep_time = 3
        if max_sec > 120:
            sleep_time = 5

        elapsed_sec = 0
        datasets = None
        while elapsed_sec < max_sec:
            time.sleep(sleep_time)
            elapsed_sec += sleep_time
            if dataset:
                datasets = conn.list_dataset_names(crawler_db)
                if (crawler_db + '.' + dataset) in datasets:
                    return datasets
            else:
                params = conn.execute_ddl('DESCRIBE DATABASE %s' % crawler_db)
                for p in params:
                    if p[0] == CRAWLER_STATUS_KEY and p[1] == 'COMPLETED':
                        return conn.list_dataset_names(crawler_db)
        if not datasets:
            datasets = conn.list_dataset_names(crawler_db)
        print("Found datasets: " + str(datasets))
        self.fail('Crawler did not find dataset in time.')

    def create_crawler_and_verify(self, user, crawler, root_path, found_tables, max_sec=25):
        crawler_db = '_okera_crawler_' + crawler
        ctx = common.get_test_context()

        with common.get_planner(ctx) as conn:
            # Always drop the crawler as admin
            ctx.disable_auth()
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % crawler_db)

            if user is not None:
                ctx.enable_token_auth(token_str=user)

            conn.execute_ddl("CREATE CRAWLER %s SOURCE '%s'" % (crawler, root_path))
            for table in found_tables:
                print("Waiting on table: %s" % table)
                self._wait_for_crawler(conn, crawler_db, table, max_sec)

    def run_crawler_test(self, user=None, config=None):
        if not config:
            return
        if config.provider in CRAWLER_TEST_MODES:
            self.create_crawler_and_verify(user,
                                           config.name,
                                           config.root_path,
                                           config.found_tables,
                                           config.timeout)

    @unittest.skipIf(common.TEST_LEVEL in ["smoke", "dev"], "Skipping at unit test level")
    def test_basic(self):
        tests = [
            CrawlerTest(provider='S3',
                        name='integration_test',
                        root_path='s3://cerebrodata-test/tpch-nation-integration-test/',
                        found_tables=['tpch_nation_integration_test'],
                        timeout=25),
            CrawlerTest(provider='ADLS',
                        name='integration_test',
                        root_path='adl://okeratestdata.azuredatalakestore.net/cerebrodata-test/tpch-nation-integration-test/',
                        found_tables=['tpch_nation_integration_test'],
                        timeout=25),
            CrawlerTest(provider='ABFS',
                        name='integration_test',
                        root_path='abfs://cerebrodata-test@okeratestdata.dfs.core.windows.net/tpch-nation-integration-test/',
                        found_tables=['tpch_nation_integration_test'],
                        timeout=25)
                ]
        for test in tests:
            self.run_crawler_test(test)

    @unittest.skipIf(('S3' not in CRAWLER_TEST_MODES),
                     "Skipping UT because platform not set in CRAWLER_TEST_MODES")
    @unittest.skipIf(common.TEST_LEVEL in ["smoke", "dev"], "Skipping at unit test level")
    def test_not_slow_5k(self):
        # This table has 5K partitions, make sure it finishes reasonably quickly
        tests = [
            CrawlerTest(provider='S3',
                        name='integration_test',
                        root_path='s3://cerebrodata-test/part_flat_5000_data_100/',
                        found_tables=['part_flat_5000_data_100'],
                        timeout=10)
                ]
        for test in tests:
            self.run_crawler_test(test)

    @unittest.skipIf(common.TEST_LEVEL in ["smoke", "dev"], "Skipping at unit test level")
    def test_flat(self):
        tests = [
            CrawlerTest(provider='S3',
                        name='integration_test',
                        root_path='s3://cerebrodata-test/crawler_flat_structure/',
                        found_tables=['atp_csv', 'topbabynamesbystate_csv'],
                        timeout=25),
            CrawlerTest(provider='ADLS',
                        name='integration_test',
                        root_path='adl://okeratestdata.azuredatalakestore.net/cerebrodata-test/crawler_flat_structure/',
                        found_tables=['atp_csv', 'topbabynamesbystate_csv'],
                        timeout=25),
            CrawlerTest(provider='ABFS',
                        name='integration_test',
                        root_path='abfs://cerebrodata-test@okeratestdata.dfs.core.windows.net/crawler_flat_structure/',
                        found_tables=['atp_csv', 'topbabynamesbystate_csv'],
                        timeout=25)
                ]
        for test in tests:
            if test.provider not in CRAWLER_TEST_MODES:
                continue
            crawler_db = '_okera_crawler_' + test.name
            ctx = common.get_test_context()
            with common.get_planner(ctx) as conn:
                conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % crawler_db)
                conn.execute_ddl("CREATE CRAWLER %s SOURCE '%s'" % (test.name, test.root_path))

                # This should only find one dataset since flat structures is not the default
                self._wait_for_crawler(conn, crawler_db, None)
                datasets = conn.list_dataset_names(crawler_db)
                self.assertEqual(len(datasets), 1)
                conn.execute_ddl('DROP TABLE %s.%s' % (crawler_db, 'crawler_flat_structure'))

                # Configure this crawler to accept flat structures
                conn.execute_ddl("ALTER DATABASE %s SET DBPROPERTIES('%s'='%s')" %\
                    (crawler_db, 'okera.crawler.single_file_datasets', 'true'))
                conn.execute_ddl("ALTER DATABASE %s SET DBPROPERTIES('%s'='%s')" %\
                    (crawler_db, CRAWLER_STATUS_KEY, ''))

                # Crawler should have been retriggered, wait for it, it should now find
                # both datasets
                for table in test.found_tables:
                    self._wait_for_crawler(conn, crawler_db, table)

    @unittest.skipIf(common.TEST_LEVEL in ["smoke", "dev"], "Skipping at unit test level")
    def test_empty(self):
        crawler = 'test_empty'
        crawler_db = '_okera_crawler_' + crawler
        if 'S3' in CRAWLER_TEST_MODES:
            location = 's3://cerebrodata-test/empty/'
        elif 'ADLS' in CRAWLER_TEST_MODES:
            location = ('adl://okeratestdata.azuredatalakestore.net/' +
                        'cerebrodata-test/empty_folder/')
        elif 'ABFS' in CRAWLER_TEST_MODES:
            location = ('abfs://cerebrodata-test@okeratestdata.dfs.core.windows.net/' +
                        'empty_folder/')
        else:
            self.fail('Test received invalid testType: ' + CRAWLER_TEST_MODES)
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % crawler_db)
            conn.execute_ddl("CREATE CRAWLER %s SOURCE '%s'" % (crawler, location))
            contents = self._wait_for_crawler(conn, crawler_db, None)

        if contents:
            self.fail('Directory was not empty, location : ' + location)

    @unittest.skipIf(common.TEST_LEVEL in ["smoke", "dev"], "Skipping at unit test level")
    def test_hidden_file(self):
        crawler = 'test_hidden_file'
        crawler_db = '_okera_crawler_' + crawler
        if 'S3' in CRAWLER_TEST_MODES:
            location = 's3://cerebrodata-test/hidden_files_only/'
        elif 'ADLS' in CRAWLER_TEST_MODES:
            location = ('adl://okeratestdata.azuredatalakestore.net/' +
                        'cerebrodata-test/hidden_file/')
        elif 'ABFS' in CRAWLER_TEST_MODES:
            location = ('abfs://cerebrodata-test@okeratestdata.dfs.core.windows.net/' +
                        'hidden_files_only/')
        else:
            self.fail('Test received invalid testType: ' + CRAWLER_TEST_MODES)
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % crawler_db)
            conn.execute_ddl("CREATE CRAWLER %s SOURCE '%s'" % (crawler, location))
            contents = self._wait_for_crawler(conn, crawler_db, None)

        if contents:
            self.fail('Directory was not empty or hidden file found, location : '
                      + location)

    @unittest.skipIf(common.TEST_LEVEL in ["smoke", "dev"], "Skipping at unit test level")
    def test_bad_path_good_scheme(self):
        caught = None
        crawler = 'test_bad_path_good_scheme'
        crawler_db = '_okera_crawler_' + crawler
        bucket = 'thispathdoesnotexistcerebro'
        if 'S3' in CRAWLER_TEST_MODES:
            location = ('s3://%s/' % bucket)
        elif 'ADLS' in CRAWLER_TEST_MODES:
            location = ('adl://%s.azuredatalakestore.net/'% bucket)
        elif 'ABFS' in CRAWLER_TEST_MODES:
            location = 'abfs://%s@okeratestdata.dfs.core.windows.net/' % bucket
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            try:
                conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % crawler_db)
                conn.execute_ddl("CREATE CRAWLER %s SOURCE '%s'" % (crawler, location))
            except Exception as e:
                if "is not accessible" not in str(e) and "does not exist" not in str(e):
                    self.fail("Exception did not match expected. Ex encountered: "
                              + str(e))
                else:
                    caught = True
            if(not caught):
                self.fail("No exception raised, expected TRecordServiceException")

    @unittest.skipIf(common.TEST_LEVEL in ["smoke", "dev"], "Skipping at unit test level")
    def test_null_path(self):
        crawler = 'test_null_path'
        caught = None
        crawler_db = '_okera_crawler' + crawler
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % crawler_db)
            try:
                conn.execute_ddl("CREATE CRAWLER %s SOURCE '%s'" % (crawler, None))
            except Exception as e:
                if "Source path not valid" not in str(e):
                    self.fail("Exception did not match expected. Ex encountered: "
                              + str(e))
                else:
                    caught = True

            if not caught:
                self.fail("No exception raised, expected TRecordServiceException")

    @unittest.skipIf(common.TEST_LEVEL in ["smoke", "dev"], "Skipping at unit test level")
    def test_bad_scheme(self):
        caught = None
        crawler = 'test_bad_scheme'
        crawler_db = '_okera_crawler' + crawler
        location = 'foo://bucket/path'
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % crawler_db)
            try:
                conn.execute_ddl("CREATE CRAWLER %s SOURCE '%s'" % (crawler, location))
            except Exception as e:
                if "Crawlers are not supported for this storage system" not in str(e):
                    self.fail("Exception did not match expected. Ex encountered: "
                              + str(e))
                else:
                    caught = True
            if not caught:
                self.fail("No exception raised, expected TRecordServiceException")

    @unittest.skipIf(common.TEST_LEVEL in ["smoke", "dev"], "Skipping at unit test level")
    def test_no_scheme(self):
        caught = None
        crawler = 'test_no_scheme'
        crawler_db = '_okera_crawler' + crawler
        location = 'bucket/path'
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % crawler_db)
            try:
                conn.execute_ddl("CREATE CRAWLER %s SOURCE '%s'" % (crawler, location))
            except Exception as e:
                if "Please verify the URL is correct:" not in str(e):
                    self.fail("Exception did not match expected. Ex encountered: "
                              + str(e))
                else:
                    caught = True
            if not caught:
                self.fail("No exception raised, expected TRecordServiceException")

    @unittest.skipIf(common.TEST_LEVEL in ["smoke", "dev"], "Skipping at unit test level")
    def test_crawl_partitioned(self):
        crawler = 'test_crawl_partitioned'
        crawler_db = '_okera_crawler_' + crawler
        if 'S3' in CRAWLER_TEST_MODES:
            location = 's3://cerebrodata-test/partitions-test2/'
        elif 'ADLS' in CRAWLER_TEST_MODES:
            location = ('adl://okeratestdata.azuredatalakestore.net/' +
                        'cerebrodata-test/partitions-test/')
        elif 'ABFS' in CRAWLER_TEST_MODES:
            location = ('abfs://cerebrodata-test@okeratestdata.dfs.core.windows.net/' +
                        'partitions-test2/')
        else:
            self.fail('Test received invalid testType: ' + CRAWLER_TEST_MODES)
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % crawler_db)
            conn.execute_ddl("CREATE CRAWLER %s SOURCE '%s'" % (crawler, location))
            contents = self._wait_for_crawler(conn, crawler_db, None, 30)

        if len(contents) != 2:
            self.fail("Did not find 2 datasets at location: " + location)

    @unittest.skipIf(common.TEST_LEVEL in ["smoke", "dev"], "Skipping at unit test level")
    def test_crawl_already_registered(self, testType=None):
        crawler = 'test_crawl_already_registered'
        crawler_db = '_okera_crawler_' + crawler
        if 'S3' in CRAWLER_TEST_MODES:
            location = 's3://cerebrodata-test/tpch_nation/'
        elif 'ADLS' in CRAWLER_TEST_MODES:
            location = ('adl://okeratestdata.azuredatalakestore.net/'
                        'cerebrodata-test/tpch_nation/')
        elif 'ABFS' in CRAWLER_TEST_MODES:
            location = ('abfs://cerebrodata-test@okeratestdata.dfs.core.windows.net/' +
                        'tpch_nation/')
        else:
            self.fail('Test received invalid testType: ' + testType)
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            if 'S3' in CRAWLER_TEST_MODES:
                conn.execute_ddl("CREATE EXTERNAL TABLE IF NOT EXISTS"
                                 "`default`.`tpch_nation_s3`(s smallint"
                                 ", n string, t smallint, d string) LOCATION '%s'"
                                 % location)
            elif 'ADLS' in CRAWLER_TEST_MODES:
                conn.execute_ddl("CREATE EXTERNAL TABLE IF NOT EXISTS"
                                 "`default`.`tpch_nation_adls`(s smallint"
                                 ", n string, t smallint, d string) LOCATION '%s'"
                                 % location)
            elif 'ABFS' in CRAWLER_TEST_MODES:
                conn.execute_ddl("CREATE EXTERNAL TABLE IF NOT EXISTS"
                                 "`default`.`tpch_nation_abfs`(s smallint"
                                 ", n string, t smallint, d string) LOCATION '%s'"
                                 % location)
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % crawler_db)
            conn.execute_ddl("CREATE CRAWLER %s SOURCE '%s'" % (crawler, location))
            contents = self._wait_for_crawler(conn, crawler_db, None)

        if contents:
            print(str(contents))
            self.fail("Found unregistered dataset when should have found none: " +
                      location)

    @unittest.skipIf(common.TEST_LEVEL in ["smoke", "dev"], "Skipping at unit test level")
    def test_crawl_duplicate_paths(self):
        crawler1 = 'test_crawl_duplicate_paths_1'
        crawler1_db = '_okera_crawler_' + crawler1
        crawler2 = 'test_crawl_duplicate_paths_2'
        crawler2_db = '_okera_crawler_' + crawler2
        if 'S3' in CRAWLER_TEST_MODES:
            location = 's3://cerebrodata-test/alltypes-parquet/'
        elif 'ADLS' in CRAWLER_TEST_MODES:
            location = ('adl://okeratestdata.azuredatalakestore.net/'
                        'cerebrodata-test/alltypes-parquet/')
        elif 'ABFS' in CRAWLER_TEST_MODES:
            location = ('abfs://cerebrodata-test@okeratestdata.dfs.core.windows.net/' +
                        'alltypes-parquet/')
        else:
            self.fail('Test received invalid testType: ' + CRAWLER_TEST_MODES)
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % crawler1_db)
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % crawler2_db)
            conn.execute_ddl("CREATE CRAWLER %s SOURCE '%s'" % (crawler1, location))
            contents1 = self._wait_for_crawler(conn, crawler1_db, None)
            conn.execute_ddl("CREATE CRAWLER %s SOURCE '%s'" % (crawler2, location))
            contents2 = self._wait_for_crawler(conn, crawler2_db, None)

            contents1 = [ds.split(".")[1] for ds in contents1]
            contents2 = [ds.split(".")[1] for ds in contents2]

        self.assertEqual(contents1, contents2)

    @staticmethod
    def __get_columns(dataset):
        cols = []
        partition_cols = []
        for c in dataset.schema.cols:
            if c.hidden:
                continue
            if c.is_partitioning_col:
                partition_cols.append(c.name)
            else:
                cols.append(c.name)
        return cols, partition_cols

    @unittest.skipIf(common.TEST_LEVEL in ["smoke", "dev"], "Skipping at unit test level")
    @unittest.skipIf(('S3' not in CRAWLER_TEST_MODES),
                     "Skipping UT because platform not set in CRAWLER_TEST_MODES")
    def test_data_partition_col_dup(self):
        # This dataset has the same column name in the partition path and data file.
        # We want to ignore the column in the data file.
        crawler = 'data_partition_col_dup_test'
        crawler_db = '_okera_crawler_' + crawler
        location = 's3://cerebro-test-itay/lake4/partitioned2/'

        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            # Default crawler allows duplicates
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % crawler_db)
            conn.execute_ddl("CREATE CRAWLER %s SOURCE '%s'" % (crawler, location))
            self._wait_for_crawler(conn, crawler_db, 'partitioned2')
            dataset = conn.list_datasets(crawler_db, name='partitioned2')
            cols, partition_cols = self.__get_columns(dataset[0])
            self.assertTrue('str2' in cols)
            self.assertTrue('str2' in partition_cols)
            show = conn.execute_ddl(
                'show create table %s.%s' % (crawler_db, 'partitioned2'))[0][0]
            self.assertEqual(2, show.count('str2 STRING'), msg=show)

            # Configure this crawler to ignore duplicates
            conn.execute_ddl("DROP TABLE %s.%s" % (crawler_db, 'partitioned2'))
            conn.execute_ddl("ALTER DATABASE %s SET DBPROPERTIES('%s'='%s')" %\
                (crawler_db, 'okera.allow.partition_cols_in_data', 'true'))
            conn.execute_ddl("ALTER DATABASE %s SET DBPROPERTIES('%s'='%s')" %\
                (crawler_db, CRAWLER_STATUS_KEY, ''))
            self._wait_for_crawler(conn, crawler_db, 'partitioned2')
            dataset = conn.list_datasets(crawler_db, name='partitioned2')
            cols, partition_cols = self.__get_columns(dataset[0])

            # str2 should not be in cols now, only in partition cols
            self.assertTrue('str2' not in cols)
            self.assertTrue('str2' in partition_cols)
            show = conn.execute_ddl(
                'show create table %s.%s' % (crawler_db, 'partitioned2'))[0][0]
            self.assertEqual(1, show.count('str2 STRING'), msg=show)

    @unittest.skipIf(common.TEST_LEVEL in ["smoke", "dev"], "Skipping at unit test level")
    @unittest.skipIf(('S3' not in CRAWLER_TEST_MODES),
                     "Skipping UT because platform not set in CRAWLER_TEST_MODES")
    def test_crawler_no_suffix(self):
        # Crawl a directory where the datasets don't have file suffixes and make
        # sure we can find them another way.
        crawler = 'no_suffix_test'
        crawler_db = '_okera_crawler_' + crawler
        location = 's3://okera-ui/autotagger/no-postfix'

        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % crawler_db)
            conn.execute_ddl("CREATE CRAWLER %s SOURCE '%s'" % (crawler, location))
            self._wait_for_crawler(conn, crawler_db)

            dataset = conn.list_datasets(crawler_db, name='csv')[0]
            self.assertEqual("TEXT", dataset.table_format)
            dataset = conn.list_datasets(crawler_db, name='parquet')[0]
            self.assertEqual("PARQUET", dataset.table_format)
            dataset = conn.list_datasets(crawler_db, name='avro')[0]
            self.assertEqual("AVRO", dataset.table_format)
            dataset = conn.list_datasets(crawler_db, name='json')[0]
            self.assertEqual("JSON", dataset.table_format)

    # DAS-5753: Find out why this is causing failures in subsequent crawler tests
    @unittest.skipIf(True, "Skipping disabled test")
    @unittest.skipIf(common.TEST_LEVEL in ["smoke", "dev"], "Skipping at unit test level")
    @unittest.skipIf(('S3' not in CRAWLER_TEST_MODES),
                     "Skipping UT because platform not set in CRAWLER_TEST_MODES")
    def test_crawler_opencsv(self):
        # Crawl a directory where the datasets don't have file suffixes and make
        # sure we can find them another way.
        crawler = 'nfl_test'
        crawler_db = '_okera_crawler_' + crawler
        location = 's3://okera-cust-success/nflstatistics/'

        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % crawler_db)
            conn.execute_ddl("CREATE CRAWLER %s SOURCE '%s'" % (crawler, location))
            self._wait_for_crawler(conn, crawler_db, 'basic_stats', max_sec=30)
            dataset = conn.list_datasets(crawler_db, name='basic_stats')[0]
            self.assertEqual(",", dataset.serde_metadata['separatorChar'])

    @unittest.skipIf(common.TEST_LEVEL in ["smoke", "dev"], "Skipping at unit test level")
    @unittest.skipIf(('S3' not in CRAWLER_TEST_MODES),
                     "Skipping UT because platform not set in CRAWLER_TEST_MODES")
    def test_crawler_tsv(self):
        # Crawl a directory where the datasets don't have file suffixes and make
        # sure we can find them another way.
        crawler = 'tsv_test'
        crawler_db = '_okera_crawler_' + crawler
        location = 's3://cerebro-test-kevin/dea_pain_pills/'

        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % crawler_db)
            conn.execute_ddl("CREATE CRAWLER %s SOURCE '%s'" % (crawler, location))
            self._wait_for_crawler(conn, crawler_db)

            dataset = conn.list_datasets(crawler_db, name='dea_pain_pills')[0]
            self.assertEqual("TEXT", dataset.table_format)
            self.assertEqual(42, len(dataset.schema.cols))

    # DAS-5753: fix and re-enable this test
    @unittest.skipIf(True, "Skipping disabled test.")
    @unittest.skipIf(common.TEST_LEVEL != "ci" and
                     common.TEST_LEVEL != "all", "Only running at CI/ALL level.")
    @unittest.skipIf(('S3' not in CRAWLER_TEST_MODES),
                     "Skipping UT because platform not set in CRAWLER_TEST_MODES")
    def test_root_crawler(self):
        # Crawl the root test bucket, this can take a while.
        crawler = 'root_crawler_test'
        crawler_db = '_okera_crawler_' + crawler
        location = 's3://cerebrodata-test/'

        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % crawler_db)
            conn.execute_ddl("CREATE CRAWLER %s SOURCE '%s'" % (crawler, location))
            # This takes ~330s on an ec2 dev box
            self._wait_for_crawler(conn, crawler_db, 'zookeeper', max_sec=600)
            datasets = conn.list_dataset_names(crawler_db)
            self.assertTrue(len(datasets) > 250, msg=str(datasets))

    @unittest.skipIf(common.TEST_LEVEL != "ci" and
                     common.TEST_LEVEL != "all", "Only running at CI/ALL level.")
    @unittest.skipIf(('ABFS' not in CRAWLER_TEST_MODES),
                     "Skipping UT because platform not set in CRAWLER_TEST_MODES")
    def test_root_crawler_abfs(self):
        # Crawl the root test bucket, this can take a while.
        crawler = 'abfs_test'
        crawler_db = '_okera_crawler_' + crawler
        location = 'abfs://cerebrodata-test@okeratestdata.dfs.core.windows.net/'

        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % crawler_db)
            conn.execute_ddl("CREATE CRAWLER %s SOURCE '%s'" % (crawler, location))
            self._wait_for_crawler(conn, crawler_db, max_sec=600)
            datasets = conn.list_dataset_names(crawler_db)
            self.assertTrue(len(datasets) > 250, msg=str(datasets))

    @unittest.skipIf(common.TEST_LEVEL != "ci" and
                     common.TEST_LEVEL != "all", "Only running at CI/ALL level.")
    @unittest.skipIf(('ADLS' not in CRAWLER_TEST_MODES),
                     "Skipping UT because platform not set in CRAWLER_TEST_MODES")
    def test_root_crawler_adls(self):
        # Crawl the root test bucket, this can take a while.
        crawler = 'adls_test'
        crawler_db = '_okera_crawler_' + crawler
        location = 'adl://okeratestdata.azuredatalakestore.net/cerebrodata-test/'

        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % crawler_db)
            conn.execute_ddl("CREATE CRAWLER %s SOURCE '%s'" % (crawler, location))
            self._wait_for_crawler(conn, crawler_db, max_sec=600)
            datasets = conn.list_dataset_names(crawler_db)
            self.assertTrue(len(datasets) > 250, msg=str(datasets))


    # DAS-5753: fix and re-enable this test
    @unittest.skipIf(True, "Skipping disabled test.")
    @unittest.skipIf(common.TEST_LEVEL != "ci" and
                     common.TEST_LEVEL != "all", "Only running at CI/ALL level.")
    @unittest.skipIf(('S3' not in CRAWLER_TEST_MODES),
                     "Skipping UT because platform not set in CRAWLER_TEST_MODES")
    def test_ui_crawler(self):
        # Crawl the ui test bucket, this can take a while.
        crawler = 'ui_test'
        crawler_db = '_okera_crawler_' + crawler
        location = 's3://okera-ui/'

        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % crawler_db)
            conn.execute_ddl("CREATE CRAWLER %s SOURCE '%s'" % (crawler, location))
            self._wait_for_crawler(conn, crawler_db, max_sec=300)
            datasets = conn.list_dataset_names(crawler_db)
            self.assertTrue(len(datasets) > 80)


    @unittest.skipIf(common.TEST_LEVEL in ["smoke", "dev"], "Skipping at unit test level")
    @unittest.skipIf(('S3' not in CRAWLER_TEST_MODES),
                     "Skipping UT because platform not set in CRAWLER_TEST_MODES")
    @unittest.skipIf(True, "Skipping as the backend broke this test - it cannot pass")
    def test_okera_lite(self):
        # This crawler is special and registers the tables in another DB based on
        # the path.
        crawler = 'okera_lite'
        result_db = 'okera_lite_default'
        crawler_db = '_okera_crawler_' + crawler
        location = 's3://cerebrodata-test/tpch-nation/'
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % crawler_db)
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % result_db)
            conn.execute_ddl("CREATE CRAWLER %s SOURCE '%s'" % (crawler, location))
            self._wait_for_crawler(conn, crawler_db, None)
            contents = conn.list_dataset_names(result_db)
            self.assertTrue('okera_lite_default.tpch_nation' in contents,
                            msg=str(contents))

    @unittest.skipIf(common.TEST_LEVEL in ["smoke", "dev"], "Skipping at unit test level")
    @unittest.skipIf(('S3' not in CRAWLER_TEST_MODES),
                     "Skipping UT because platform not set in CRAWLER_TEST_MODES")
    def test_das_3489(self):
        crawler = 'das_3489_test'
        crawler_db = '_okera_crawler_' + crawler
        location = 's3://cerebrodata-test/put-test2/'
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % crawler_db)
            conn.execute_ddl("CREATE CRAWLER %s SOURCE '%s'" % (crawler, location))
            self._wait_for_crawler(conn, crawler_db, max_sec=300)
            dataset = conn.list_datasets(crawler_db, name='put_test2')[0]
            self.assertEqual("TEXT", dataset.table_format)

    @unittest.skipIf(common.TEST_LEVEL in ["smoke", "dev"], "Skipping at unit test level")
    @unittest.skipIf(('S3' not in CRAWLER_TEST_MODES),
                     "Skipping UT because platform not set in CRAWLER_TEST_MODES")
    def test_empty_json(self):
        # This is an empty json file. We should not register it
        crawler = 'json_test'
        crawler_db = '_okera_crawler_' + crawler
        location = 's3://cerebro-datasets/empty-json/empty.json'
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % crawler_db)
            conn.execute_ddl("CREATE CRAWLER %s SOURCE '%s'" % (crawler, location))
            self._wait_for_crawler(conn, crawler_db, max_sec=10)
            self.assertEqual(len(conn.list_datasets(crawler_db)), 0)

    @unittest.skipIf(common.TEST_LEVEL in ["smoke", "dev"], "Skipping at unit test level")
    @unittest.skipIf(('S3' not in CRAWLER_TEST_MODES),
                     "Skipping UT because platform not set in CRAWLER_TEST_MODES")
    def test_text_boolean(self):
        # Test case for a cedilla delimited file
        crawler = 'text_boolean_test'
        crawler_db = '_okera_crawler_' + crawler
        location = 's3://okera-demo/consent-management/whitelist'
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % crawler_db)
            conn.execute_ddl("CREATE CRAWLER %s SOURCE '%s'" % (crawler, location))
            self._wait_for_crawler(conn, crawler_db, max_sec=10)
            dataset = conn.list_datasets(crawler_db, name='whitelist')[0]
            self.assertEqual(3, len(dataset.schema.cols))
            self.assertEqual(
                _thrift_api.TTypeId.STRING,
                dataset.schema.cols[0].type.type_id)
            self.assertEqual(
                _thrift_api.TTypeId.BOOLEAN,
                dataset.schema.cols[1].type.type_id)
            self.assertEqual(
                _thrift_api.TTypeId.BOOLEAN,
                dataset.schema.cols[2].type.type_id)

    @unittest.skipIf(common.TEST_LEVEL in ["smoke", "dev"], "Skipping at unit test level")
    @unittest.skipIf(('S3' not in CRAWLER_TEST_MODES),
                     "Skipping UT because platform not set in CRAWLER_TEST_MODES")
    def test_cedilla_delimiter(self):
        # Test case for a cedilla delimited file
        crawler = 'cedilla_delimiter_test'
        crawler_db = '_okera_crawler_' + crawler
        location = 's3://cerebro-datasets/cedilla_sample/'
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % crawler_db)
            conn.execute_ddl("CREATE CRAWLER %s SOURCE '%s'" % (crawler, location))
            self._wait_for_crawler(conn, crawler_db, max_sec=10)
            dataset = conn.list_datasets(crawler_db, name='cedilla_sample')[0]
            self.assertEqual(2, len(dataset.schema.cols))

    @unittest.skipIf(common.TEST_LEVEL in ["smoke", "dev"], "Skipping at unit test level")
    @unittest.skipIf(('S3' not in CRAWLER_TEST_MODES),
                     "Skipping UT because platform not set in CRAWLER_TEST_MODES")
    def test_comments(self):
        crawler = 'comments_test'
        crawler_db = '_okera_crawler_' + crawler
        location = 's3://cerebrodata-test/crawler-schema-comments/'
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % crawler_db)
            conn.execute_ddl("CREATE CRAWLER %s SOURCE '%s'" % (crawler, location))
            self._wait_for_crawler(conn, crawler_db)

            # This should have comments from the data file
            whitelist = conn.list_datasets(crawler_db, name='avro_whitelist')[0]
            self.assertTrue('This dataset stores' in whitelist.description,
                            msg=str(whitelist))
            whitelist_col = whitelist.schema.cols[1]
            self.assertTrue('Indicates if the user' in whitelist_col.comment,
                            msg=str(whitelist_col))

            # This should have default comments
            default_comments = conn.list_datasets(crawler_db, name='avro_no_comments')[0]
            self.assertTrue('Discovered by Okera crawler' in default_comments.description,
                            msg=str(default_comments))
            default_comments_col = default_comments.schema.cols[1]
            self.assertTrue(default_comments_col.comment is None)

    @unittest.skipIf(common.TEST_LEVEL in ["smoke", "dev"], "Skipping at unit test level")
    @unittest.skipIf(('S3' not in CRAWLER_TEST_MODES),
                     "Skipping UT because platform not set in CRAWLER_TEST_MODES")
    @unittest.skipIf(('AUTOTAGGER' not in CRAWLER_TEST_MODES),
                     "Skipping UT because Autotagger not set in CRAWLER_TEST_MODES")
    def test_autotagger(self):
        crawler = 'autotagger_test'
        crawler_db = '_okera_crawler_' + crawler
        location = 's3://okera-datalake/gdpr-avro/transactions/'
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            # Run it a few times to make sure dropping table and assigning existing
            # tags works.
            for _ in range(3):
                conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % crawler_db)
                conn.execute_ddl("CREATE CRAWLER %s SOURCE '%s'" % (crawler, location))
                self._wait_for_crawler(conn, crawler_db)

                # This should have comments from the data file
                transactions = conn.list_datasets(crawler_db, name='transactions')[0]
                self.assertTrue('This dataset stores' in transactions.description,
                                msg=str(transactions))

                # This column should be tagged
                ccn = transactions.schema.cols[5]
                self.assertTrue(ccn.attribute_values is not None)
                self.assertEqual('credit_card_number',
                                 ccn.attribute_values[0].attribute.key)
                self.assertTrue('credit card number' in ccn.comment,
                                msg=str(ccn))

    @unittest.skipIf(common.TEST_LEVEL in ["smoke", "dev"], "Skipping at unit test level")
    def test_crawl_ordering(self):
        crawler = 'test_crawl_ordering'
        crawler_db = '_okera_crawler_' + crawler
        if 'S3' in CRAWLER_TEST_MODES:
            location = 's3://cerebrodata-test/partitioned_flat_50000/'
        elif 'ADLS' in CRAWLER_TEST_MODES:
            location = ('adl://okeratestdata.azuredatalakestore.net/' +
                        'cerebrodata-test/partitioned_flat_50000/')
        elif 'ABFS' in CRAWLER_TEST_MODES:
            location = ('abfs://cerebrodata-test@okeratestdata.dfs.core.windows.net/' +
                        'partitioned_flat_50000/')
        else:
            self.fail('Test received invalid testType: ' + CRAWLER_TEST_MODES)
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % crawler_db)
            conn.execute_ddl("CREATE CRAWLER %s SOURCE '%s'" % (crawler, location))
            contents = self._wait_for_crawler(conn, crawler_db, None, 30)

            self.assertEqual(
                contents,
                ['_okera_crawler_test_crawl_ordering.partitioned_flat_50000'])
            dataset = conn.list_datasets(crawler_db, name='partitioned_flat_50000')[0]
            datafile_path = dataset.metadata[CRAWLER_DATA_FILE_PATH_KEY]
            self.assertEqual(
                datafile_path,
                's3a://cerebrodata-test/partitioned_flat_50000/part=9999/data_0.csv')

    @unittest.skipIf(common.TEST_LEVEL in ["smoke", "dev"], "Skipping at unit test level")
    def test_crawl_dup_names(self):
        crawler = 'test_crawl_dup_names'
        crawler_db = '_okera_crawler_' + crawler
        base_loc = "cerebrodata-test/crawler_dupe_names/"
        if 'S3' in CRAWLER_TEST_MODES:
            location = "s3://" + base_loc
        elif 'ADLS' in CRAWLER_TEST_MODES:
            location = "adl://okeratestdata.azuredatalakestore.net/" + base_loc
        elif 'ABFS' in CRAWLER_TEST_MODES:
            location = ("abfs://cerebrodata-test@okeratestdata.dfs.core.windows.net/" +
                        base_loc)
        else:
            self.fail("Test received invalid testType: " + CRAWLER_TEST_MODES)

        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % crawler_db)
            conn.execute_ddl("CREATE CRAWLER %s SOURCE '%s'" % (crawler, location))
            contents = self._wait_for_crawler(conn, crawler_db, None, 30)

            self.assertIn('_okera_crawler_test_crawl_dup_names.transactions_1', contents)

    @unittest.skipIf(common.TEST_LEVEL in ["smoke", "dev"], "Skipping at unit test level")
    def test_das_4807(self):
        # This test verifies that we are picking up the latest file in the folder which
        # contains data with quotes, which should produce a different result in the
        # crawled table. This may not be robust as this is a "real" dataset that gets
        # data added to it. If this because they problem, we should create a directory
        # with 2 test files, the older with no quotes and the newer with quotes.
        crawler = 'test_crawl_health'
        crawler_db = '_okera_crawler_' + crawler
        location = 's3://cerebro-test-kevin/covid_19_daily'

        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % crawler_db)
            conn.execute_ddl("CREATE CRAWLER %s SOURCE '%s'" % (crawler, location))
            contents = self._wait_for_crawler(conn, crawler_db, None, 30)
            self.assertIn('_okera_crawler_test_crawl_health.covid_19_daily', contents)
            dataset = conn.list_datasets(crawler_db, name='covid_19_daily')[0]
            self.assertEqual(
                "org.apache.hadoop.hive.serde2.lazy.LazySimpleSerDe", dataset.serialization_lib)

    @unittest.skipIf(common.TEST_LEVEL in ["smoke", "dev"], "Skipping at unit test level")
    def test_col_name_too_long(self):
        crawler = 'test_crawl_long_col_name'
        crawler_db = '_okera_crawler_' + crawler
        if 'S3' in CRAWLER_TEST_MODES:
            location = 's3://cerebrodata-test/long-col-names/'
        elif 'ADLS' in CRAWLER_TEST_MODES:
            location = ('adl://okeratestdata.azuredatalakestore.net/'
                        'cerebrodata-test/long-col-names/')
        elif 'ABFS' in CRAWLER_TEST_MODES:
            location = ('abfs://cerebrodata-test@okeratestdata.dfs.core.windows.net/' +
                        'long-col-names/')
        else:
            self.fail('Test received invalid testType: ' + CRAWLER_TEST_MODES)

        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % crawler_db)
            conn.execute_ddl("CREATE CRAWLER %s SOURCE '%s'" % (crawler, location))
            self._wait_for_crawler(conn, crawler_db, max_sec=10)
            self.assertEqual(len(conn.list_datasets(crawler_db)), 0)

    def test_crawler_privileges(self):
        s3_uri = 's3://cerebrodata-test/tpch-nation-integration-test/'
        crawler_name = 'integration_test'
        table_name = 'tpch_nation_integration_test'
        as_owner_role = "_okera_internal_role_%s" % (TEST_USER)

        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:

            for priv in [PRIV_ALL, PRIV_SHOW, PRIV_SELECT, PRIV_DROP, PRIV_ALTER]:
                ctx.disable_auth()
                setup_ddls = [
                    # Setup create role
                    "DROP ROLE IF EXISTS %s" % (TEST_ROLE),
                    "CREATE ROLE %s" % (TEST_ROLE),
                    "GRANT ROLE %s TO GROUP %s" % (TEST_ROLE, TEST_USER),

                    # Setup use role
                    "DROP ROLE IF EXISTS %s" % (TEST_ROLE2),
                    "CREATE ROLE %s" % (TEST_ROLE2),
                    "GRANT ROLE %s TO GROUP %s" % (TEST_ROLE2, TEST_USER2),

                    # Cleanup the special per-user role to start from a clean slate
                    "DROP ROLE IF EXISTS %s" % (as_owner_role),

                    # Grant create_crawler_as_owner on the catalog and access to the
                    # URI
                    "GRANT CREATE_CRAWLER_AS_OWNER ON CATALOG TO ROLE %s" % (TEST_ROLE),
                    "GRANT ALL ON URI '%s' TO ROLE %s" % (s3_uri, TEST_ROLE),
                ]

                for ddl in setup_ddls:
                    conn.execute_ddl(ddl)

                tests = [
                    CrawlerTest(provider='S3',
                                name=crawler_name,
                                root_path=s3_uri,
                                found_tables=[table_name],
                                timeout=25),
                ]
                for test in tests:
                    self.run_crawler_test(TEST_USER, test)

                user2_ddls = [
                    # Setup use role
                    "DROP ROLE IF EXISTS %s" % (TEST_ROLE2),
                    "CREATE ROLE %s" % (TEST_ROLE2),
                    "GRANT ROLE %s TO GROUP %s" % (TEST_ROLE2, TEST_USER2),

                    # Grant create_crawler_as_owner on the catalog and access to the
                    # URI
                    "GRANT %s ON CRAWLER %s TO ROLE %s" % (priv.level, crawler_name, TEST_ROLE2),
                ]
                for ddl in user2_ddls:
                    conn.execute_ddl(ddl)

                # Switch to second user who has access to just this crawler
                ctx.enable_token_auth(token_str=TEST_USER2)

                print("Running crawler privilege tests with privilege: %s" % priv.level)

                # Ensure we can "wait" for the crawler
                crawler_db = '_okera_crawler_' + crawler_name
                self._wait_for_crawler(conn, crawler_db, table_name, 25)

                # Ensure we can read the data or not
                query = "SELECT * FROM %s.%s WHERE n_name='ALGERIA'" % (crawler_db, table_name)
                if priv.can_select:
                    res = conn.scan_as_json(query)

                    assert len(res) == 1
                    assert res[0]['n_nationkey'] == 0
                else:
                    with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                        conn.scan_as_json(query)
                    self.assertTrue("'testcrawleruser2' does not have privileges to execute 'SELECT'" in str(ex_ctx.exception))

                ddl = "ALTER DATABASE %s SET dbproperties('okera.crawler.status'='')" % crawler_db
                if priv.can_alter:
                    conn.execute_ddl(ddl)
                else:
                    with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                        conn.execute_ddl(ddl)
                    self.assertTrue("'testcrawleruser2' does not have privileges to execute 'ALTER'" in str(ex_ctx.exception))

                ddl = "DROP DATABASE %s CASCADE" % crawler_db
                if priv.can_drop:
                    conn.execute_ddl(ddl)
                else:
                    with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                        conn.execute_ddl(ddl)
                    self.assertTrue("'testcrawleruser2' does not have privileges to execute 'DROP'" in str(ex_ctx.exception))


if __name__ == "__main__":
    unittest.main()



import os

from .. import context as base_context

__DOMINO_TOKEN_FILE_ENV = 'DOMINO_TOKEN_FILE'
__DOMINO_DEFAULT_USER_CLAIM = 'preferred_username'

def token_func():
    if __DOMINO_TOKEN_FILE_ENV not in os.environ:
        raise Exception('Could not find Domino token environment variable: %s' % (
            __DOMINO_TOKEN_FILE_ENV))

    token_path = os.environ[__DOMINO_TOKEN_FILE_ENV]
    if not os.path.exists(token_path):
        raise Exception('Could not find Domino token file at "%s"' % token_path)

    return open(token_path, 'r').read().strip()

def context(*args, user_claims=None, **kwargs):
    """ Gets a pyokera context object configured for Domino Data Labs environments.

        Parameters
        ----------
        user_claims : list of str, optional
            Optional list of claims to extract user from token

    Returns
    -------
    OkeraContext
        Context object.

    Examples
    --------
    >>> from okera.integration import domino
    >>> ctx = domino.context()
    >>> with ctx.connect(host = 'okera.company.com', port = 12050) as conn:
    ...     result = conn.execute_ddl('select * from okera_sample.whoami')
    ...     result
    [{'user': '<domino user>'}]
    """
    # If there was no explicit user_claims set, we use the default one for
    # Domino
    if not user_claims:
        user_claims = [__DOMINO_DEFAULT_USER_CLAIM]
    ctx = base_context(*args, **kwargs)
    ctx.enable_token_auth(token_func=token_func, user_claims=user_claims)

    return ctx
